@@ -11,7 +11,7 @@
 
 QT_BEGIN_NAMESPACE
 
-Q_LOGGING_CATEGORY(lcFolderListModel, "qt.labs.folderlistmodel")
+Q_STATIC_LOGGING_CATEGORY(lcFolderListModel, "qt.labs.folderlistmodel")
 
 class QQuickFolderListModelPrivate
 {
@@ -226,7 +226,7 @@ QString QQuickFolderListModelPrivate::resolvePath(const QUrl &path)
 /*!
     \qmltype FolderListModel
     \inqmlmodule Qt.labs.folderlistmodel
-//!    \instantiates QQuickFolderListModel
+//!    \nativetype QQuickFolderListModel
     \ingroup qtquick-models
     \brief The FolderListModel provides a model of the contents of a file system folder.
 
