@@ -9,7 +9,6 @@
 #include <qv4scopedvalue_p.h>
 #include <qv4runtime_p.h>
 #include <qv4variantobject_p.h>
-#include "qv4string_p.h"
 #include "qv4jscall_p.h"
 #include <qv4symbol_p.h>
 
@@ -115,6 +114,7 @@ QChar JsonParser::nextToken()
     case EndArray:
     case EndObject:
         eatSpace();
+        break;
     case Quote:
         break;
     default:
@@ -623,19 +623,20 @@ public:
                         QStringLiteral("Cannot convert circular structure to JSON"));
         }
 
-        stringify->v4->checkStackLimits();
+        const bool hasOverflow = stringify->v4->checkStackLimits();
+        Q_UNUSED(hasOverflow); // No handling needed. Exception has been thrown already.
     }
 
     bool foundProblem() const { return m_callDepthRecorder.ee->hasException; }
 
 private:
-    ExecutionEngineCallDepthRecorder m_callDepthRecorder;
+    ExecutionEngineCallDepthRecorder<1> m_callDepthRecorder;
 };
 
 static QString quote(const QString &str)
 {
     QString product;
-    const int length = str.length();
+    const int length = str.size();
     product.reserve(length + 2);
     product += u'"';
     for (int i = 0; i < length; ++i) {
@@ -696,12 +697,18 @@ QString Stringify::Str(const QString &key, const Value &v)
     }
 
     if (replacerFunction) {
-        ScopedObject holder(scope, v4->newObject());
-        holder->put(scope.engine->id_empty(), value);
         JSCallArguments jsCallData(scope, 2);
         jsCallData.args[0] = v4->newString(key);
         jsCallData.args[1] = value;
-        *jsCallData.thisObject = holder;
+
+        if (stack.isEmpty()) {
+            ScopedObject holder(scope, v4->newObject());
+            holder->put(scope.engine->id_empty(), v);
+            *jsCallData.thisObject = holder;
+        } else {
+            *jsCallData.thisObject = stack.top();
+        }
+
         value = replacerFunction->call(jsCallData);
         if (v4->hasException)
             return QString();
@@ -885,7 +892,7 @@ ReturnedValue JsonObject::method_parse(const FunctionObject *b, const Value *, c
         jtext = argv[0].toQString();
 
     DEBUG << "parsing source = " << jtext;
-    JsonParser parser(v4, jtext.constData(), jtext.length());
+    JsonParser parser(v4, jtext.constData(), jtext.size());
     QJsonParseError error;
     ReturnedValue result = parser.parse(&error);
     if (error.error != QJsonParseError::NoError) {
@@ -982,12 +989,16 @@ QJsonValue JsonObject::toJsonValue(const Value &value, V4ObjectSet &visitedObjec
 
     Q_ASSERT(value.isObject());
     Scope scope(value.as<Object>()->engine());
-    ScopedArrayObject a(scope, value);
-    if (a)
+    if (ScopedArrayObject a{ scope, value }) {
         return toJsonArray(a, visitedObjects);
-    ScopedObject o(scope, value);
-    if (o)
+    } else if (Scoped<QV4::Sequence> a{ scope, value }) {
+        return toJsonArray(a, visitedObjects);
+    } else if (Scoped<QmlListWrapper> lw{ scope, value }) {
+        return toJsonArray(lw, visitedObjects);
+    } else if (ScopedObject o{ scope, value }) {
         return toJsonObject(o, visitedObjects);
+    }
+
     return QJsonValue(value.toQString());
 }
 
@@ -1052,7 +1063,7 @@ QV4::ReturnedValue JsonObject::fromJsonArray(ExecutionEngine *engine, const QJso
     return a.asReturnedValue();
 }
 
-QJsonArray JsonObject::toJsonArray(const ArrayObject *a, V4ObjectSet &visitedObjects)
+QJsonArray JsonObject::toJsonArray(const Object *a, V4ObjectSet &visitedObjects)
 {
     QJsonArray result;
     if (!a)
