@@ -131,9 +131,9 @@ PropertyAttributes StringObject::virtualGetOwnProperty(const Managed *m, Propert
 
 DEFINE_OBJECT_VTABLE(StringCtor);
 
-void Heap::StringCtor::init(QV4::ExecutionContext *scope)
+void Heap::StringCtor::init(QV4::ExecutionEngine *engine)
 {
-    Heap::FunctionObject::init(scope, QStringLiteral("String"));
+    Heap::FunctionObject::init(engine, QStringLiteral("String"));
 }
 
 ReturnedValue StringCtor::virtualCallAsConstructor(const FunctionObject *f, const Value *argv, int argc, const Value *newTarget)
@@ -778,7 +778,14 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
             nMatchOffsets += re->captureCount() * 2;
             if (!regExp->global())
                 break;
-            offset = qMax(offset + 1, matchOffsets[oldSize + 1]);
+
+            const uint matchBegin = matchOffsets[oldSize];
+            const uint matchEnd = matchOffsets[oldSize + 1];
+
+            // If we have a zero-sized match, don't match at the same place again.
+            const uint matchOffset = (matchBegin == matchEnd) ? matchEnd + 1 : matchEnd;
+
+            offset = std::max(offset + 1, matchOffset);
         }
         if (regExp->global()) {
             regExp->setLastIndex(0);
@@ -1012,6 +1019,11 @@ ReturnedValue StringPrototype::method_startsWith(const FunctionObject *b, const
     if (argc > 1)
         pos = argv[1].toInteger();
 
+    pos = std::clamp(
+        pos,
+        0.0,
+        double(value.size()));
+
     if (pos == 0)
         return Encode(value.startsWith(searchString));
 
@@ -1096,9 +1108,55 @@ ReturnedValue StringPrototype::method_toLowerCase(const FunctionObject *b, const
     return Encode(v4->newString(value.toLower()));
 }
 
+static const QLocale *getLocaleDataResource(const QV4::Value &val)
+{
+    if (const QV4::QQmlValueTypeWrapper *wrapper = val.as<QQmlValueTypeWrapper>())
+        return wrapper->cast<const QLocale>();
+
+    return nullptr;
+}
+
+static QLocale getLocaleFromArgs(const FunctionObject *b, const QV4::Value *argv, int argc)
+{
+    if (argc == 0)
+        return QLocale();
+
+    ExecutionEngine *v4 = b->engine();
+    Scope scope(b);
+    QString stringifiedLocale;
+    // First argument is a string, we check if it is a valid locale
+    if (const QV4::String *that = argv[0].as<QV4::String>()) {
+        stringifiedLocale = that->toQString();
+    } else if (const QLocale *locale = getLocaleDataResource(argv[0])) {
+        return *locale;
+    } else if (argv[0].isObject()) {
+        // First argument is an array, we check if the first element is
+        // a string and a valid locale.
+        ScopedObject arrayLike(scope, argv[0].toObject(scope.engine));
+        ScopedValue kValue(scope);
+        if (arrayLike->getLength() > 0) {
+            kValue = arrayLike->get(uint(0));
+            if (kValue->isString())
+                stringifiedLocale = kValue->toQString();
+            else if (const QLocale *locale = getLocaleDataResource(kValue))
+                return *locale;
+        }
+    } else {
+        v4->throwTypeError();
+    }
+
+    return QLocale(stringifiedLocale);
+}
+
 ReturnedValue StringPrototype::method_toLocaleLowerCase(const FunctionObject *b, const Value *thisObject, const Value *argv, int argc)
 {
-    return method_toLowerCase(b, thisObject, argv, argc);
+    ExecutionEngine *v4 = b->engine();
+    const QString value = getThisString(v4, thisObject);
+    if (v4->hasException)
+        return QV4::Encode::undefined();
+
+    QLocale locale = getLocaleFromArgs(b, argv, argc);
+    return Encode(v4->newString(locale.toLower(value)));
 }
 
 ReturnedValue StringPrototype::method_toUpperCase(const FunctionObject *b, const Value *thisObject, const Value *, int)
@@ -1113,7 +1171,13 @@ ReturnedValue StringPrototype::method_toUpperCase(const FunctionObject *b, const
 
 ReturnedValue StringPrototype::method_toLocaleUpperCase(const FunctionObject *b, const Value *thisObject, const Value *argv, int argc)
 {
-    return method_toUpperCase(b, thisObject, argv, argc);
+    ExecutionEngine *v4 = b->engine();
+    const QString value = getThisString(v4, thisObject);
+    if (v4->hasException)
+        return QV4::Encode::undefined();
+
+    QLocale locale = getLocaleFromArgs(b, argv, argc);
+    return Encode(v4->newString(locale.toUpper(value)));
 }
 
 ReturnedValue StringPrototype::method_trim(const FunctionObject *b, const Value *thisObject, const Value *, int)
