@@ -36,7 +36,7 @@ namespace {
     Q_CONSTINIT thread_local int creationDepth = 0;
 }
 
-Q_LOGGING_CATEGORY(lcQmlComponentGeneral, "qt.qml.qmlcomponent")
+Q_STATIC_LOGGING_CATEGORY(lcQmlComponentGeneral, "qt.qml.qmlcomponent")
 
 QT_BEGIN_NAMESPACE
 
@@ -186,6 +186,12 @@ V4_DEFINE_EXTENSION(QQmlComponentExtension, componentExtension);
     \c Component objects can also be created dynamically using
     \l{QtQml::Qt::createComponent()}{Qt.createComponent()}.
 
+    \c {Component}s are useful to declare a type where you only need an
+    instance of the type without having to add an entire new file. However,
+    you cannot name this type and consequently can't use it to declare a
+    property or use it in a type annotation. If you need this, prefer using
+    \l{Defining Object Types through QML Documents#Inline Components}{inline components}.
+
     \section2 Creation Context
 
     The creation context of a Component corresponds to the context where the Component was declared.
@@ -322,8 +328,12 @@ void QQmlComponentPrivate::clear()
         typeData.reset();
     }
 
+    if (loadHelper) {
+        loadHelper->unregisterCallback(this);
+        loadHelper.reset();
+    }
+
     compilationUnit.reset();
-    loadedType = {};
     inlineComponentName.reset();
 }
 
@@ -356,7 +366,7 @@ static void removePendingQPropertyBinding(
     if (QQmlData *ddata = QQmlData::get(o)) {
         const QQmlPropertyData *propData = ddata->propertyCache->property(
             propertyName, o, ddata->outerContext);
-        if (propData && propData->isBindable())
+        if (propData && propData->acceptsQBinding())
             creator->removePendingBinding(o, propData->coreIndex());
         return;
     }
@@ -386,7 +396,8 @@ bool QQmlComponentPrivate::setInitialProperty(
         }
         const QString lastProperty = properties.last();
         segment = scope.engine->newString(lastProperty);
-        object->put(segment, scope.engine->metaTypeToJS(value.metaType(), value.constData()));
+        QV4::ScopedValue v(scope, scope.engine->metaTypeToJS(value.metaType(), value.constData()));
+        object->put(segment, v);
         if (scope.engine->hasException) {
             qmlWarning(base, scope.engine->catchExceptionAsQmlError());
             scope.engine->hasException = false;
@@ -459,6 +470,10 @@ QQmlComponent::~QQmlComponent()
 
     if (d->typeData) {
         d->typeData->unregisterCallback(d);
+        if (d->engine) {
+            QQmlEnginePrivate::get(d->engine)->typeLoader.drop(
+                    QQmlDataBlob::Ptr(d->typeData.data()));
+        }
         d->typeData.reset();
     }
 }
@@ -488,8 +503,10 @@ QQmlComponent::Status QQmlComponent::status() const
         return Loading;
     else if (!d->state.errors.isEmpty())
         return Error;
-    else if (d->engine && (d->compilationUnit || d->loadedType.isValid()))
+    else if (d->engine && (d->compilationUnit || d->loadedType().isValid()))
         return Ready;
+    else if (d->loadHelper)
+        return Loading;
     else
         return Null;
 }
@@ -694,6 +711,9 @@ QQmlComponent::QQmlComponent(QQmlEngine *engine, QV4::ExecutableCompilationUnit
     Sets the QQmlComponent to use the given QML \a data. If \a url
     is provided, it is used to set the component name and to provide
     a base path for items resolved by this component.
+
+    \warning The new component will shadow any existing component of
+    the same URL. You should not pass a URL of an existing component.
 */
 void QQmlComponent::setData(const QByteArray &data, const QUrl &url)
 {
@@ -1023,6 +1043,14 @@ QObject *QQmlComponent::beginCreate(QQmlContext *context)
     return d->beginCreate(QQmlContextData::get(context));
 }
 
+static QQmlParserStatus *parserStatusCast(const QQmlType &type, QObject *rv)
+{
+    const int parserStatusCast = type.parserStatusCast();
+    return parserStatusCast == -1
+            ? nullptr
+            : reinterpret_cast<QQmlParserStatus *>(reinterpret_cast<char *>(rv) + parserStatusCast);
+}
+
 QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> context)
 {
     Q_Q(QQmlComponent);
@@ -1078,12 +1106,14 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
 
     QObject *rv = nullptr;
 
-    if (!loadedType.isValid()) {
+    const QQmlType type = loadedType();
+    if (!type.isValid()) {
         enginePriv->referenceScarceResources();
-        state.initCreator(std::move(context), compilationUnit, creationContext);
+        const QString *icName = inlineComponentName.get();
+        state.initCreator(context, compilationUnit, creationContext, icName ? *icName : QString());
 
         QQmlObjectCreator::CreationFlags flags;
-        if (const QString *icName = inlineComponentName.get()) {
+        if (icName) {
             flags = QQmlObjectCreator::InlineComponent;
             if (start == -1)
                 start = compilationUnit->inlineComponentId(*icName);
@@ -1098,28 +1128,23 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
         enginePriv->dereferenceScarceResources();
     } else {
         // TODO: extract into function
-        rv = loadedType.createWithQQmlData();
+        rv = type.createWithQQmlData();
         QQmlPropertyCache::ConstPtr propertyCache = QQmlData::ensurePropertyCache(rv);
-        QQmlParserStatus *parserStatus = nullptr;
-        const int parserStatusCast = loadedType.parserStatusCast();
-        if (parserStatusCast != -1) {
-            parserStatus = reinterpret_cast<QQmlParserStatus*>(reinterpret_cast<char *>(rv) + parserStatusCast);
+        if (QQmlParserStatus *parserStatus = parserStatusCast(type, rv)) {
             parserStatus->classBegin();
+            state.ensureRequiredPropertyStorage(rv);
+        } else if (type.finalizerCast() != -1) {
+            state.ensureRequiredPropertyStorage(rv);
         }
+
         for (int i = 0, propertyCount = propertyCache->propertyCount(); i < propertyCount; ++i) {
             if (const QQmlPropertyData *propertyData = propertyCache->property(i); propertyData->isRequired()) {
-                state.ensureRequiredPropertyStorage();
+                state.ensureRequiredPropertyStorage(rv);
                 RequiredPropertyInfo info;
                 info.propertyName = propertyData->name(rv);
                 state.addPendingRequiredProperty(rv, propertyData, info);
             }
         }
-        if (parserStatus)
-            parserStatus->componentComplete();
-        if (const int finalizerCast = loadedType.finalizerCast(); finalizerCast != -1) {
-            auto* hook = reinterpret_cast<QQmlFinalizerHook *>(reinterpret_cast<char *>(rv) + finalizerCast);
-            hook->componentFinalized();
-        }
     }
 
     if (rv) {
@@ -1130,6 +1155,12 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
         ddata->indestructible = true;
         ddata->explicitIndestructibleSet = true;
         ddata->rootObjectInCreation = false;
+
+        // Assign parent context to the object if we haven't created one.
+        if (!ddata->outerContext)
+            ddata->outerContext = context.data();
+        if (!ddata->context)
+            ddata->context = context.data();
     }
 
     return rv;
@@ -1152,7 +1183,9 @@ void QQmlComponentPrivate::beginDeferred(QQmlEnginePrivate *enginePriv,
         auto creator = state.initCreator(
                     deferredData->context->parent(),
                     deferredData->compilationUnit,
-                    QQmlRefPointer<QQmlContextData>());
+                    QQmlRefPointer<QQmlContextData>(),
+                    deferredData->inlineComponentName
+                );
 
         if (!creator->populateDeferredProperties(object, deferredData))
             state.appendCreatorErrors();
@@ -1266,7 +1299,19 @@ void QQmlComponentPrivate::completeCreate()
             state.errors.push_back(QQmlComponentPrivate::AnnotatedQmlError { error, true });
         }
     }
-    if (loadedType.isValid()) {
+
+    const QQmlType type = loadedType();
+    if (type.isValid()) {
+        QObject *rv = state.target();
+        if (QQmlParserStatus *parserStatus = parserStatusCast(type, rv))
+            parserStatus->componentComplete();
+
+        if (const int finalizerCast = type.finalizerCast(); finalizerCast != -1) {
+            auto *hook = reinterpret_cast<QQmlFinalizerHook *>(
+                    reinterpret_cast<char *>(rv) + finalizerCast);
+            hook->componentFinalized();
+        }
+
         /*
            We can directly set completePending to false, as finalize is only concerned
            with setting up pending bindings, but that cannot happen here, as we're
@@ -1340,23 +1385,37 @@ void QQmlComponent::loadFromModule(QAnyStringView uri, QAnyStringView typeName,
                                    QQmlComponent::CompilationMode mode)
 {
     Q_D(QQmlComponent);
-    auto [status, type] = d->prepareLoadFromModule(uri, typeName);
-    d->completeLoadFromModule(uri, typeName, type, status, mode);
+
+    QQmlTypeLoader::Mode typeLoaderMode = QQmlTypeLoader::Synchronous;
+    switch (mode) {
+    case QQmlComponent::PreferSynchronous:
+        typeLoaderMode = QQmlTypeLoader::PreferSynchronous;
+        break;
+    case QQmlComponent::Asynchronous:
+        typeLoaderMode = QQmlTypeLoader::Asynchronous;
+        break;
+    }
+
+    d->prepareLoadFromModule(uri, typeName, typeLoaderMode);
+    if (d->loadHelper->isCompleteOrError())
+        d->completeLoadFromModule(uri, typeName);
+    else
+        d->loadHelper->registerCallback(d);
 }
 
-LoadHelper::ResolveTypeResult QQmlComponentPrivate::prepareLoadFromModule(QAnyStringView uri,
-                                                                          QAnyStringView typeName)
+void QQmlComponentPrivate::prepareLoadFromModule(
+        QAnyStringView uri, QAnyStringView typeName, QQmlTypeLoader::Mode mode)
 {
-    auto enginePriv = QQmlEnginePrivate::get(engine);
-    // LoadHelper must be on the Heap as it derives from QQmlRefCount
-    auto loadHelper = QQml::makeRefPointer<LoadHelper>(&enginePriv->typeLoader, uri);
+    // Don't let any old loadHelper call us back anymore.
+    if (loadHelper)
+        loadHelper->unregisterCallback(this);
 
-    return loadHelper->resolveType(typeName);
+    QQmlTypeLoader *typeLoader = &QQmlEnginePrivate::get(engine)->typeLoader;
+    // LoadHelper must be on the Heap as it derives from QQmlRefCount
+    loadHelper = QQml::makeRefPointer<LoadHelper>(typeLoader, uri, typeName, mode);
 }
 
-void QQmlComponentPrivate::completeLoadFromModule(QAnyStringView uri, QAnyStringView typeName, QQmlType type,
-                                                  LoadHelper::ResolveTypeResult::Status moduleStatus,
-                                                  QQmlComponent::CompilationMode mode)
+void QQmlComponentPrivate::completeLoadFromModule(QAnyStringView uri, QAnyStringView typeName)
 {
     Q_Q(QQmlComponent);
 
@@ -1381,16 +1440,28 @@ void QQmlComponentPrivate::completeLoadFromModule(QAnyStringView uri, QAnyString
         emit q->statusChanged(q->status());
     };
     emitProgressReset();
-    if (moduleStatus == LoadHelper::ResolveTypeResult::NoSuchModule) {
+
+    const QQmlType type = loadHelper->type();
+
+    if (loadHelper->resolveTypeResult() == LoadHelper::ResolveTypeResult::NoSuchModule)  {
         reportError(QLatin1String(R"(No module named "%1" found)").arg(uri.toString()));
     } else if (!type.isValid()) {
         reportError(QLatin1String(R"(Module "%1" contains no type named "%2")")
                     .arg(uri.toString(), typeName.toString()));
     } else if (type.isCreatable()) {
-        clear();
-        loadedType = type;
         emitComplete();
     } else if (type.isComposite()) {
+        QQmlComponent::CompilationMode mode = QQmlComponent::PreferSynchronous;
+        switch (loadHelper->mode()) {
+        case QQmlTypeLoader::Asynchronous:
+            mode = QQmlComponent::Asynchronous;
+            break;
+        case QQmlTypeLoader::PreferSynchronous:
+        case QQmlTypeLoader::Synchronous:
+            mode = QQmlComponent::PreferSynchronous;
+            break;
+        }
+
         // loadUrl takes care of signal emission
         loadUrl(type.sourceUrl(), mode);
     } else if (type.isInlineComponentType()) {
@@ -1476,7 +1547,7 @@ void QQmlComponent::create(QQmlIncubator &incubator, QQmlContext *context, QQmlC
     incubator.clear();
     QExplicitlySharedDataPointer<QQmlIncubatorPrivate> p(incubator.d);
 
-    if (d->loadedType.isValid()) {
+    if (d->loadedType().isValid()) {
         // there isn't really an incubation process for C++ backed types
         // so just create the object and signal that we are ready
 
@@ -1488,7 +1559,9 @@ void QQmlComponent::create(QQmlIncubator &incubator, QQmlContext *context, QQmlC
 
     p->compilationUnit = d->compilationUnit;
     p->enginePriv = enginePriv;
-    p->creator.reset(new QQmlObjectCreator(contextData, d->compilationUnit, d->creationContext, p.data()));
+    p->creator.reset(new QQmlObjectCreator(contextData, d->compilationUnit, d->creationContext,
+                                           d->inlineComponentName ? *d->inlineComponentName : QString(),
+                                           p.data()));
     p->subComponentToCreate = d->start;
 
     enginePriv->incubate(incubator, forContextData);
@@ -1560,7 +1633,7 @@ void QQmlComponentPrivate::incubateObject(
 
     incubatorPriv->compilationUnit = componentPriv->compilationUnit;
     incubatorPriv->enginePriv = enginePriv;
-    incubatorPriv->creator.reset(new QQmlObjectCreator(context, componentPriv->compilationUnit, componentPriv->creationContext));
+    incubatorPriv->creator.reset(new QQmlObjectCreator(context, componentPriv->compilationUnit, componentPriv->creationContext, inlineComponentName ? *inlineComponentName : QString()));
 
     if (start == -1) {
         if (const QString *icName = componentPriv->inlineComponentName.get()) {
