@@ -1,23 +1,24 @@
 // Copyright (C) 2022 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
-#include <qglobal.h>
-
+#include <QtCore/qcompilerdetection.h>
 // GCC 11 thinks diagMsg.fixSuggestion.fixes.d.ptr is somehow uninitialized in
 // QList::emplaceBack(), probably called from QQmlJsLogger::log()
 // Ditto for GCC 12, but it emits a different warning
 QT_WARNING_PUSH
 QT_WARNING_DISABLE_GCC("-Wuninitialized")
 QT_WARNING_DISABLE_GCC("-Wmaybe-uninitialized")
-#include <qlist.h>
+#include <QtCore/qlist.h>
 QT_WARNING_POP
 
-#include "qqmljslogger_p.h"
-#include "qqmljsloggingutils.h"
-#include "qqmlsa_p.h"
+#include <private/qqmljslogger_p.h>
+#include <private/qqmlsa_p.h>
 
+#include <QtQmlCompiler/qqmljsloggingutils.h>
+
+#include <QtCore/qglobal.h>
 #include <QtCore/qfile.h>
-#include <QtCore/qfileinfo.h>
+
 
 QT_BEGIN_NAMESPACE
 
@@ -182,10 +183,13 @@ const QList<QQmlJS::LoggerCategory> &QQmlJSLogger::defaultCategories()
                 QStringLiteral("Warn if a singleton is accessed via an object"), QtWarningMsg },
         QQmlJS::LoggerCategory{
                 qmlTopLevelComponent.name().toString(), QStringLiteral("TopLevelComponent"),
-                QStringLiteral("Fail when a top level Component are encountered"), QtWarningMsg },
+                QStringLiteral("Warn if a top level Component is encountered"), QtWarningMsg },
         QQmlJS::LoggerCategory{
                 qmlUncreatableType.name().toString(), QStringLiteral("UncreatableType"),
-                QStringLiteral("Warn if uncreatable types are created"), QtWarningMsg }
+                QStringLiteral("Warn if uncreatable types are created"), QtWarningMsg },
+        QQmlJS::LoggerCategory{
+                qmlMissingEnumEntry.name().toString(), QStringLiteral("MissingEnumEntry"),
+                QStringLiteral("Warn about using missing enum values"), QtWarningMsg },
     };
 
     return cats;
@@ -248,15 +252,14 @@ void QQmlJSLogger::log(const QString &message, QQmlJS::LoggerWarningId id,
 
     QString prefix;
 
-    if ((!overrideFileName.isEmpty() || !m_fileName.isEmpty()) && showFileName)
+    if ((!overrideFileName.isEmpty() || !m_filePath.isEmpty()) && showFileName)
         prefix =
-                (!overrideFileName.isEmpty() ? overrideFileName : m_fileName) + QStringLiteral(":");
+                (!overrideFileName.isEmpty() ? overrideFileName : m_filePath) + QStringLiteral(":");
 
     if (srcLocation.isValid())
-        prefix += QStringLiteral("%1:%2:").arg(srcLocation.startLine).arg(srcLocation.startColumn);
-
-    if (!prefix.isEmpty())
-        prefix.append(QLatin1Char(' '));
+        prefix += QStringLiteral("%1:%2: ").arg(srcLocation.startLine).arg(srcLocation.startColumn);
+    else if (!prefix.isEmpty())
+        prefix += QStringLiteral(": "); // produce double colon for Qt Creator's issues pane
 
     // Note: we do the clamping to [Info, Critical] range since our logger only
     // supports 3 categories
@@ -273,11 +276,11 @@ void QQmlJSLogger::log(const QString &message, QQmlJS::LoggerWarningId id,
     diagMsg.type = type;
     diagMsg.fixSuggestion = suggestion;
 
-    switch (type) {
-    case QtWarningMsg: m_warnings.push_back(diagMsg); break;
-    case QtCriticalMsg: m_errors.push_back(diagMsg); break;
-    case QtInfoMsg: m_infos.push_back(diagMsg); break;
-    default: break;
+    if (m_inTransaction) {
+        m_pendingMessages.push_back(std::move(diagMsg));
+    } else {
+        countMessage(diagMsg);
+        m_committedMessages.push_back(std::move(diagMsg));
     }
 
     if (srcLocation.length > 0 && !m_code.isEmpty() && showContext)
@@ -285,10 +288,28 @@ void QQmlJSLogger::log(const QString &message, QQmlJS::LoggerWarningId id,
 
     if (suggestion.has_value())
         printFix(suggestion.value());
+
+    if (!m_inTransaction)
+        m_output.flushBuffer();
+}
+
+void QQmlJSLogger::countMessage(const Message &message)
+{
+    switch (message.type) {
+    case QtWarningMsg:
+        ++m_numWarnings;
+        break;
+    case QtCriticalMsg:
+        ++m_numErrors;
+        break;
+    default:
+        break;
+    }
 }
 
 void QQmlJSLogger::processMessages(const QList<QQmlJS::DiagnosticMessage> &messages,
-                                   QQmlJS::LoggerWarningId id)
+                                   QQmlJS::LoggerWarningId id,
+                                   const QQmlJS::SourceLocation &sourceLocation)
 {
     if (messages.isEmpty() || isCategoryIgnored(id))
         return;
@@ -298,17 +319,63 @@ void QQmlJSLogger::processMessages(const QList<QQmlJS::DiagnosticMessage> &messa
     // TODO: we should instead respect message's category here (potentially, it
     // should hold a category instead of type)
     for (const QQmlJS::DiagnosticMessage &message : messages)
-        log(message.message, id, QQmlJS::SourceLocation(), false, false);
+        log(message.message, id, sourceLocation, false, false);
 
     m_output.write(QStringLiteral("---\n\n"));
 }
 
+/*!
+    \internal
+    Starts a transaction for a compile pass. This buffers all messages until the
+    transaction completes. If you commit the transaction, the messages are printed
+    and added to the list of committed messages. If you roll it back, the logger
+    reverts to the state before the start of the transaction.
+
+    This is useful for compile passes that potentially have to be repeated, such
+    as the type propagator. We don't want to see the same messages logged multiple
+    times.
+ */
+void QQmlJSLogger::startTransaction()
+{
+    Q_ASSERT(!m_inTransaction);
+    m_inTransaction = true;
+}
+
+/*!
+    \internal
+    Commit the current transaction. Print all pending messages, and add them to
+    the list of committed messages. Then, clear the transaction flag.
+ */
+void QQmlJSLogger::commit()
+{
+    Q_ASSERT(m_inTransaction);
+    for (const Message &message : std::as_const(m_pendingMessages))
+        countMessage(message);
+
+    m_committedMessages.append(std::exchange(m_pendingMessages, {}));
+    m_output.flushBuffer();
+    m_inTransaction = false;
+}
+
+/*!
+    \internal
+    Roll back the current transaction and revert the logger to the state before
+    it was started.
+ */
+void QQmlJSLogger::rollback()
+{
+    Q_ASSERT(m_inTransaction);
+    m_pendingMessages.clear();
+    m_output.discardBuffer();
+    m_inTransaction = false;
+}
+
 void QQmlJSLogger::printContext(const QString &overrideFileName,
                                 const QQmlJS::SourceLocation &location)
 {
     QString code = m_code;
 
-    if (!overrideFileName.isEmpty() && overrideFileName != QFileInfo(m_fileName).absolutePath()) {
+    if (!overrideFileName.isEmpty() && overrideFileName != m_filePath) {
         QFile file(overrideFileName);
         const bool success = file.open(QFile::ReadOnly);
         Q_ASSERT(success);
@@ -339,7 +406,7 @@ void QQmlJSLogger::printContext(const QString &overrideFileName,
 
 void QQmlJSLogger::printFix(const QQmlJSFixSuggestion &fixItem)
 {
-    const QString currentFileAbsPath = QFileInfo(m_fileName).absolutePath();
+    const QString currentFileAbsPath = m_filePath;
     QString code = m_code;
     QString currentFile;
     m_output.writePrefixedMessage(fixItem.fixDescription(), QtInfoMsg);
