@@ -1,5 +1,6 @@
 // Copyright (C) 2016 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
+// Qt-Security score:critical reason:data-parser
 
 #include "qquicktextcontrol_p.h"
 #include "qquicktextcontrol_p_p.h"
@@ -42,7 +43,6 @@
 const int textCursorWidth = 1;
 
 QT_BEGIN_NAMESPACE
-Q_DECLARE_LOGGING_CATEGORY(lcHoverTrace)
 
 // could go into QTextCursor...
 static QTextLine currentTextLine(const QTextCursor &cursor)
@@ -82,6 +82,7 @@ QQuickTextControlPrivate::QQuickTextControlPrivate()
       hoveredMarker(false),
       selectByTouchDrag(false),
       imSelectionAfterPress(false),
+      beingEdited(false),
       lastSelectionStart(-1),
       lastSelectionEnd(-1)
 {}
@@ -269,7 +270,7 @@ void QQuickTextControlPrivate::setContent(Qt::TextFormat format, const QString &
     const int oldCursorPos = cursor.position();
 
     // avoid multiple textChanged() signals being emitted
-    qmlobject_disconnect(doc, QTextDocument, SIGNAL(contentsChanged()), q, QQuickTextControl, SIGNAL(textChanged()));
+    QObject::disconnect(doc, &QTextDocument::contentsChanged, q, &QQuickTextControl::textChanged);
 
     if (!text.isEmpty()) {
         // clear 'our' cursor for insertion to prevent
@@ -309,7 +310,7 @@ void QQuickTextControlPrivate::setContent(Qt::TextFormat format, const QString &
     }
     cursor.setCharFormat(charFormatForInsertion);
 
-    qmlobject_connect(doc, QTextDocument, SIGNAL(contentsChanged()), q, QQuickTextControl, SIGNAL(textChanged()));
+    QObject::connect(doc, &QTextDocument::contentsChanged, q, &QQuickTextControl::textChanged);
     emit q->textChanged();
     doc->setUndoRedoEnabled(previousUndoRedoState);
     _q_updateCurrentCharFormatAndSelection();
@@ -584,25 +585,7 @@ void QQuickTextControl::clear()
 QQuickTextControl::QQuickTextControl(QTextDocument *doc, QObject *parent)
     : QInputControl(TextEdit, *new QQuickTextControlPrivate, parent)
 {
-    Q_D(QQuickTextControl);
-    Q_ASSERT(doc);
-
-    QAbstractTextDocumentLayout *layout = doc->documentLayout();
-    qmlobject_connect(layout, QAbstractTextDocumentLayout, SIGNAL(update(QRectF)), this, QQuickTextControl, SIGNAL(updateRequest()));
-    qmlobject_connect(layout, QAbstractTextDocumentLayout, SIGNAL(updateBlock(QTextBlock)), this, QQuickTextControl, SIGNAL(updateRequest()));
-    qmlobject_connect(doc, QTextDocument, SIGNAL(contentsChanged()), this, QQuickTextControl, SIGNAL(textChanged()));
-    qmlobject_connect(doc, QTextDocument, SIGNAL(contentsChanged()), this, QQuickTextControl, SLOT(_q_updateCurrentCharFormatAndSelection()));
-    qmlobject_connect(doc, QTextDocument, SIGNAL(cursorPositionChanged(QTextCursor)), this, QQuickTextControl, SLOT(_q_updateCursorPosChanged(QTextCursor)));
-    connect(doc, &QTextDocument::contentsChange, this, &QQuickTextControl::contentsChange);
-
-    layout->setProperty("cursorWidth", textCursorWidth);
-
-    d->doc = doc;
-    d->cursor = QTextCursor(doc);
-    d->lastCharFormat = d->cursor.charFormat();
-    doc->setPageSize(QSizeF(0, 0));
-    doc->setModified(false);
-    doc->setUndoRedoEnabled(true);
+    setDocument(doc);
 }
 
 QQuickTextControl::~QQuickTextControl()
@@ -618,7 +601,32 @@ QTextDocument *QQuickTextControl::document() const
 void QQuickTextControl::setDocument(QTextDocument *doc)
 {
     Q_D(QQuickTextControl);
+    if (!doc || d->doc == doc)
+        return;
+
+    if (d->doc) {
+        QAbstractTextDocumentLayout *oldLayout = d->doc->documentLayout();
+        disconnect(oldLayout, nullptr, this, nullptr);
+        disconnect(d->doc, nullptr, this, nullptr);
+    }
+
     d->doc = doc;
+    d->cursor = QTextCursor(doc);
+    d->lastCharFormat = d->cursor.charFormat();
+
+    QAbstractTextDocumentLayout *layout = doc->documentLayout();
+    connect(layout, &QAbstractTextDocumentLayout::update, this, &QQuickTextControl::updateRequest);
+    connect(layout, &QAbstractTextDocumentLayout::updateBlock, this, &QQuickTextControl::updateRequest);
+    connect(doc, &QTextDocument::contentsChanged, this, [d]() {
+        d->_q_updateCurrentCharFormatAndSelection();
+    });
+    connect(doc, &QTextDocument::contentsChanged, this, &QQuickTextControl::textChanged);
+    connect(doc, &QTextDocument::cursorPositionChanged, this, [d](const QTextCursor &cursor) {
+        d->_q_updateCursorPosChanged(cursor);
+    });
+    connect(doc, &QTextDocument::contentsChange, this, &QQuickTextControl::contentsChange);
+    if (auto *qtdlayout = qobject_cast<QTextDocumentLayout *>(layout))
+        qtdlayout->setCursorWidth(textCursorWidth);
 }
 
 void QQuickTextControl::updateCursorRectangle(bool force)
@@ -762,6 +770,12 @@ bool QQuickTextControl::event(QEvent *e)
     return QObject::event(e);
 }
 
+bool QQuickTextControl::isBeingEdited()
+{
+    Q_D(QQuickTextControl);
+    return d->beingEdited;
+}
+
 void QQuickTextControl::timerEvent(QTimerEvent *e)
 {
     Q_D(QQuickTextControl);
@@ -800,6 +814,7 @@ void QQuickTextControlPrivate::keyPressEvent(QKeyEvent *e)
 {
     Q_Q(QQuickTextControl);
 
+    QScopedValueRollback<bool> rollbackBeingEdited(beingEdited, true);
     if (e->key() == Qt::Key_Back) {
          e->ignore();
          return;
