@@ -9,6 +9,7 @@
 #include <QtQuickTemplates2/private/qquickslider_p.h>
 
 #include <qpa/qplatformintegration.h>
+#include <qpa/qplatformservices.h>
 #include <private/qguiapplication_p.h>
 
 QT_BEGIN_NAMESPACE
@@ -91,7 +92,7 @@ QQuickColorDialogImplAttached *QQuickColorDialogImplPrivate::attachedOrWarn()
 
 void QQuickColorDialogImplPrivate::eyeDropperEnter()
 {
-    Q_Q(const QQuickColorDialogImpl);
+    Q_Q(QQuickColorDialogImpl);
     if (m_eyeDropperMode)
         return;
 
@@ -104,6 +105,19 @@ void QQuickColorDialogImplPrivate::eyeDropperEnter()
         m_eyeDropperWindow = window;
     }
 
+    if (auto *platformServices = QGuiApplicationPrivate::platformIntegration()->services();
+        platformServices && platformServices->hasCapability(QPlatformServices::Capability::ColorPicking)) {
+        if (auto *colorPickerService = platformServices->colorPicker(m_eyeDropperWindow)) {
+            q->connect(colorPickerService, &QPlatformServiceColorPicker::colorPicked, q,
+                       [q, colorPickerService](const QColor &color) {
+                           colorPickerService->deleteLater();
+                           q->setColor(color);
+                       });
+            colorPickerService->pickColor();
+            return;
+        }
+    }
+
     m_eyeDropperPreviousColor = q->color();
 
     if (!bool(eyeDropperEventFilter))
@@ -166,6 +180,7 @@ void QQuickColorDialogImplPrivate::alphaSliderMoved()
 QQuickColorDialogImpl::QQuickColorDialogImpl(QObject *parent)
     : QQuickDialog(*(new QQuickColorDialogImplPrivate), parent)
 {
+    setPopupType(QQuickPopup::Window);
 }
 
 QQuickColorDialogImplAttached *QQuickColorDialogImpl::qmlAttachedProperties(QObject *object)
@@ -183,7 +198,7 @@ QColor QQuickColorDialogImpl::color() const
 void QQuickColorDialogImpl::setColor(const QColor &c)
 {
     Q_D(QQuickColorDialogImpl);
-    if (color().rgba() == c.rgba())
+    if (const QColor old = color(); old.spec() == c.spec() && old == c)
         return;
 
     // If we get a QColor from an Hsv or Hsl color system,
@@ -423,9 +438,13 @@ void QQuickColorDialogImpl::setOptions(const QSharedPointer<QColorDialogOptions>
     QQuickColorDialogImplAttached *attached = d->attachedOrWarn();
 
     if (attached) {
-        const bool screenGrabbingAllowed = QGuiApplicationPrivate::platformIntegration()->hasCapability(QPlatformIntegration::ScreenWindowGrabbing);
+        const auto *integration = QGuiApplicationPrivate::platformIntegration();
+        const bool canSupportEyeDropper =
+                integration->hasCapability(QPlatformIntegration::ScreenWindowGrabbing)
+                || integration->services()->hasCapability(QPlatformServices::Capability::ColorPicking);
         const bool offscreen = qgetenv("QT_QPA_PLATFORM").compare(QLatin1String("offscreen"), Qt::CaseInsensitive) == 0;
-        attached->eyeDropperButton()->setVisible(screenGrabbingAllowed && !offscreen);
+        const bool noEyeDropperButton = (d->options && d->options->options() & QColorDialogOptions::NoEyeDropperButton);
+        attached->eyeDropperButton()->setVisible(!noEyeDropperButton && canSupportEyeDropper && !offscreen);
 
         if (d->options) {
             attached->buttonBox()->setVisible(
