@@ -155,14 +155,10 @@ void tst_qmlls_highlighting::highlights_data()
         if (!f.open(QIODevice::ReadOnly | QIODevice::Text))
             return file;
         QString code = f.readAll();
-        DomCreationOptions options;
-        options.setFlag(DomCreationOption::WithScriptExpressions);
-        options.setFlag(DomCreationOption::WithSemanticAnalysis);
-        options.setFlag(DomCreationOption::WithRecovery);
 
         QStringList dirs = {QLibraryInfo::path(QLibraryInfo::Qml2ImportsPath)};
         auto envPtr = DomEnvironment::create(
-                dirs, QQmlJS::Dom::DomEnvironment::Option::SingleThreaded, options);
+                dirs, QQmlJS::Dom::DomEnvironment::Option::SingleThreaded, Extended);
         envPtr->loadBuiltins();
         envPtr->loadFile(FileToLoad::fromMemory(envPtr, filePath, code),
                          [&file](Path, const DomItem &, const DomItem &newIt) {
@@ -476,6 +472,34 @@ void tst_qmlls_highlighting::highlights_data()
         QTest::addRow("null") << fileItem
                               << Token(QQmlJS::SourceLocation(285, 4, 12, 21),
                                        int(SemanticTokenProtocolTypes::Keyword), 0);
+        QTest::addRow("leftbacktick") << fileItem
+                                      << Token(QQmlJS::SourceLocation(390, 1, 17, 43),
+                                               int(SemanticTokenProtocolTypes::String), 0);
+        QTest::addRow("rightbacktick") << fileItem
+                                       << Token(QQmlJS::SourceLocation(424, 1, 20, 5),
+                                                int(SemanticTokenProtocolTypes::String), 0);
+        QTest::addRow("templatestringpartStart") << fileItem
+                                            << Token(QQmlJS::SourceLocation(391, 5, 17, 44),
+                                                     int(SemanticTokenProtocolTypes::String), 0);
+        QTest::addRow("templatestringpartEnd") << fileItem
+                                            << Token(QQmlJS::SourceLocation(412, 7, 19, 1),
+                                                     int(SemanticTokenProtocolTypes::String), 0);
+        QTest::addRow("templateExpressionPartB")
+                << fileItem
+                << Token(QQmlJS::SourceLocation(403, 1, 18, 7),
+                         int(SemanticTokenProtocolTypes::Property), 0);
+        QTest::addRow("templateExpressionPartK")
+                << fileItem
+                << Token(QQmlJS::SourceLocation(407, 1, 18, 11),
+                         int(SemanticTokenProtocolTypes::Method), 0);
+        QTest::addRow("dollarLeftBrace")
+                << fileItem
+                << Token(QQmlJS::SourceLocation(401, 2, 18, 5),
+                         int(SemanticTokenProtocolTypes::Operator), 0);
+        QTest::addRow("rightbrace")
+                << fileItem
+                << Token(QQmlJS::SourceLocation(410, 1, 18, 14),
+                         int(SemanticTokenProtocolTypes::Operator), 0);
     }
     { // identifiers
         const auto filePath = m_highlightingDataDir + "/Identifiers.qml";
@@ -816,5 +840,39 @@ void tst_qmlls_highlighting::computeDiff()
     }
 }
 
+void tst_qmlls_highlighting::enumCrash()
+{
+    using namespace QQmlJS::Dom;
+    const auto fileObject = [](const QString &filePath) {
+        QFile f(filePath);
+        DomItem file;
+        if (!f.open(QIODevice::ReadOnly | QIODevice::Text))
+            return file;
+        QString code = f.readAll();
+
+        QStringList dirs = { QLibraryInfo::path(QLibraryInfo::Qml2ImportsPath) };
+        auto envPtr = DomEnvironment::create(
+                dirs, QQmlJS::Dom::DomEnvironment::Option::SingleThreaded, Extended);
+        envPtr->loadBuiltins();
+        envPtr->loadFile(FileToLoad::fromMemory(envPtr, filePath, code),
+                         [&file](Path, const DomItem &, const DomItem &newIt) {
+                             file = newIt.fileObject();
+                         });
+        envPtr->loadPendingDependencies();
+        return file;
+    };
+
+    const auto filePath = m_highlightingDataDir + "/enums_qtbug.qml";
+    const auto fileItem = fileObject(filePath);
+
+    Highlights h;
+    HighlightingVisitor hv(h, std::nullopt);
+
+    fileItem.visitTree(QQmlJS::Dom::Path(), hv, VisitOption::Default, emptyChildrenVisitor,
+                       emptyChildrenVisitor);
+
+    const auto highlights = h.highlights();
+    QVERIFY(!highlights.isEmpty());
+}
 
 QTEST_MAIN(tst_qmlls_highlighting)
