@@ -1,8 +1,11 @@
 // Copyright (C) 2018 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
-#include <QtTest/QtTest>
+#include <QtTest/QTest>
+#include <QtTest/QSignalSpy>
 
+#include <QtGui/qstylehints.h>
+#include <QtQml/private/qqmlglobal_p.h>
 #include <QtQml/qqmlengine.h>
 #include <QtQml/qqmlproperty.h>
 #include <QtQuick/private/qquickdraghandler_p.h>
@@ -15,6 +18,8 @@
 #include <QtQuickTestUtils/private/qmlutils_p.h>
 #include <QtQuickTestUtils/private/viewtestutils_p.h>
 
+#include <QtCore/qpointer.h>
+
 Q_LOGGING_CATEGORY(lcPointerTests, "qt.quick.pointer.tests")
 
 class tst_DragHandler : public QQmlDataTest
@@ -47,6 +52,8 @@ private slots:
     void touchPinchAndMouseMove();
     void unsuitableEventDuringDrag();
     void underModalLayer();
+    void interruptedByIrrelevantButton();
+    void touchDragExclusiveGrabber();
 
 private:
     void sendWheelEvent(QQuickView &window, QPoint pos, QPoint angleDelta, QPoint pixelDelta, Qt::KeyboardModifiers modifiers, Qt::ScrollPhase phase, bool inverted);
@@ -139,8 +146,9 @@ void tst_DragHandler::touchDrag()
         dragHandler->setDragThreshold(dragThreshold);
     }
 
-    QSignalSpy translationChangedSpy(dragHandler, SIGNAL(translationChanged()));
+    QSignalSpy translationChangedSpy(dragHandler, &QQuickDragHandler::translationChanged);
     QSignalSpy centroidChangedSpy(dragHandler, SIGNAL(centroidChanged()));
+    QSignalSpy xDeltaSpy(dragHandler->xAxis(), &QQuickDragAxis::activeValueChanged);
 
     QPointF ballCenter = ball->clipRect().center();
     QPointF scenePressPos = ball->mapToScene(ballCenter);
@@ -153,21 +161,22 @@ void tst_DragHandler::touchDrag()
     QCOMPARE(dragHandler->centroid().scenePosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().scenePressPosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().velocity(), QVector2D());
-    QCOMPARE(centroidChangedSpy.count(), 1);
+    QCOMPARE(centroidChangedSpy.size(), 1);
     p1 += QPoint(dragThreshold, 0);
     QTest::touchEvent(window, touchDevice).move(1, p1, window);
     QQuickTouchUtils::flush(window);
     qCDebug(lcPointerTests) << "velocity after drag" << dragHandler->centroid().velocity();
     if (dragThreshold > 0)
         QTRY_VERIFY(!qFuzzyIsNull(dragHandler->centroid().velocity().x()));
-    QCOMPARE(centroidChangedSpy.count(), 2);
+    QCOMPARE(centroidChangedSpy.size(), 2);
     QVERIFY(!dragHandler->active());
     p1 += QPoint(1, 0);
     QTest::touchEvent(window, touchDevice).move(1, p1, window);
     QQuickTouchUtils::flush(window);
     QTRY_VERIFY(dragHandler->active());
-    QCOMPARE(translationChangedSpy.count(), 0);
-    QCOMPARE(centroidChangedSpy.count(), 3);
+    QCOMPARE(translationChangedSpy.size(), 0);
+    QCOMPARE(xDeltaSpy.size(), 0);
+    QCOMPARE(centroidChangedSpy.size(), 3);
     QCOMPARE(dragHandler->persistentTranslation().x(), 0);
     QCOMPARE(dragHandler->activeTranslation().x(), 0);
     QPointF sceneGrabPos = p1;
@@ -185,16 +194,20 @@ void tst_DragHandler::touchDrag()
     QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold + 20);
     QCOMPARE(dragHandler->persistentTranslation().y(), 0);
     QCOMPARE(dragHandler->activeTranslation().y(), 0);
+    QCOMPARE(translationChangedSpy.size(), 1);
+    QCOMPARE(translationChangedSpy.first().first().value<QVector2D>(), QVector2D(dragThreshold + 20, 0));
     QVERIFY(dragHandler->centroid().velocity().x() > 0);
-    QCOMPARE(centroidChangedSpy.count(), 4);
+    QCOMPARE(centroidChangedSpy.size(), 4);
     QTest::touchEvent(window, touchDevice).release(1, p1, window);
     QQuickTouchUtils::flush(window);
     QTRY_VERIFY(!dragHandler->active());
     QCOMPARE(dragHandler->centroid().pressedButtons(), Qt::NoButton);
     QCOMPARE(dragHandler->centroid().velocity(), QVector2D());
     QCOMPARE(ball->mapToScene(ballCenter).toPoint(), p1);
-    QCOMPARE(translationChangedSpy.count(), 1);
-    QCOMPARE(centroidChangedSpy.count(), 5);
+    QCOMPARE(translationChangedSpy.size(), 1);
+    QCOMPARE(xDeltaSpy.size(), 1);
+    QCOMPARE(xDeltaSpy.first().first().toReal(), dragThreshold + 20);
+    QCOMPARE(centroidChangedSpy.size(), 5);
     QCOMPARE(dragHandler->persistentTranslation().x(), dragThreshold + 20);
 
     // Drag again: activeTranslation starts over, while persistentTranslation accumulates
@@ -212,11 +225,15 @@ void tst_DragHandler::touchDrag()
     p1 += QPoint(9, 0);
     QTest::touchEvent(window, touchDevice).move(1, p1, window);
     QQuickTouchUtils::flush(window);
+    QCOMPARE(xDeltaSpy.size(), 2);
+    QCOMPARE(xDeltaSpy.last().first().toReal(), dragThreshold + 10);
     p1 += QPoint(10, 0);
     QTest::touchEvent(window, touchDevice).move(1, p1, window);
     QQuickTouchUtils::flush(window);
     QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold + 20);
     QCOMPARE(dragHandler->persistentTranslation().x(), dragThreshold * 2 + 40);
+    QCOMPARE(xDeltaSpy.size(), 3);
+    QCOMPARE(xDeltaSpy.last().first().toReal(), 10);
     QTest::touchEvent(window, touchDevice).release(1, p1, window);
     QQuickTouchUtils::flush(window);
 
@@ -242,6 +259,8 @@ void tst_DragHandler::touchDrag()
     QQuickTouchUtils::flush(window);
     QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold + 20);
     QCOMPARE(dragHandler->persistentTranslation().x(), dragThreshold + 30);
+    QCOMPARE(xDeltaSpy.size(), 6);
+    QCOMPARE(xDeltaSpy.last().first().toReal(), 10);
     QTest::touchEvent(window, touchDevice).release(1, p1, window);
     QQuickTouchUtils::flush(window);
 }
@@ -273,8 +292,9 @@ void tst_DragHandler::mouseDrag()
     QVERIFY(dragHandler);
     dragHandler->setAcceptedButtons(acceptedButtons); // QTBUG-76875
 
-    QSignalSpy translationChangedSpy(dragHandler, SIGNAL(translationChanged()));
+    QSignalSpy translationChangedSpy(dragHandler, &QQuickDragHandler::translationChanged);
     QSignalSpy centroidChangedSpy(dragHandler, SIGNAL(centroidChanged()));
+    QSignalSpy xDeltaSpy(dragHandler->xAxis(), &QQuickDragAxis::activeValueChanged);
 
     QPointF ballCenter = ball->clipRect().center();
     QPointF scenePressPos = ball->mapToScene(ballCenter);
@@ -290,13 +310,13 @@ void tst_DragHandler::mouseDrag()
         QCOMPARE(dragHandler->centroid().scenePosition(), scenePressPos);
         QCOMPARE(dragHandler->centroid().scenePressPosition(), scenePressPos);
         QCOMPARE(dragHandler->centroid().velocity(), QVector2D());
-        QCOMPARE(centroidChangedSpy.count(), 1);
+        QCOMPARE(centroidChangedSpy.size(), 1);
     }
     p1 += QPoint(dragThreshold, 0);
     QTest::mouseMove(window, p1);
     if (shouldDrag) {
-//        QTRY_VERIFY(dragHandler->centroid().velocity().x() > 0); // TODO QTBUG-33891
-        QCOMPARE(centroidChangedSpy.count(), 2);
+        QVERIFY(dragHandler->centroid().velocity().x() > 0);
+        QCOMPARE(centroidChangedSpy.size(), 2);
         QVERIFY(!dragHandler->active());
 #if QT_CONFIG(cursor)
         QCOMPARE(window->cursor().shape(), Qt::ArrowCursor);
@@ -308,9 +328,10 @@ void tst_DragHandler::mouseDrag()
         QTRY_VERIFY(dragHandler->active());
     else
         QVERIFY(!dragHandler->active());
-    QCOMPARE(translationChangedSpy.count(), 0);
+    QCOMPARE(translationChangedSpy.size(), 0);
+    QCOMPARE(xDeltaSpy.size(), 0);
     if (shouldDrag)
-        QCOMPARE(centroidChangedSpy.count(), 3);
+        QCOMPARE(centroidChangedSpy.size(), 3);
     QCOMPARE(dragHandler->persistentTranslation().x(), 0.0);
     QCOMPARE(dragHandler->activeTranslation().x(), 0.0);
     QPointF sceneGrabPos = p1;
@@ -329,8 +350,8 @@ void tst_DragHandler::mouseDrag()
         QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold + 20.0);
         QCOMPARE(dragHandler->persistentTranslation().y(), 0.0);
         QCOMPARE(dragHandler->activeTranslation().y(), 0.0);
-//        QVERIFY(dragHandler->centroid().velocity().x() > 0); // TODO QTBUG-33891
-        QCOMPARE(centroidChangedSpy.count(), 4);
+        QVERIFY(dragHandler->centroid().velocity().x() > 0);
+        QCOMPARE(centroidChangedSpy.size(), 4);
 #if QT_CONFIG(cursor)
         QCOMPARE(window->cursor().shape(), Qt::ClosedHandCursor);
 #endif
@@ -338,10 +359,13 @@ void tst_DragHandler::mouseDrag()
     QTest::mouseRelease(window, static_cast<Qt::MouseButton>(int(dragButton)), Qt::NoModifier, p1);
     QTRY_VERIFY(!dragHandler->active());
     QCOMPARE(dragHandler->centroid().pressedButtons(), Qt::NoButton);
-    if (shouldDrag)
+    QCOMPARE(translationChangedSpy.size(), shouldDrag ? 1 : 0);
+    if (shouldDrag) {
         QCOMPARE(ball->mapToScene(ballCenter).toPoint(), p1);
-    QCOMPARE(translationChangedSpy.count(), shouldDrag ? 1 : 0);
-    QCOMPARE(centroidChangedSpy.count(), shouldDrag ? 5 : 0);
+        QCOMPARE(translationChangedSpy.first().first().value<QVector2D>(), QVector2D(dragThreshold + 20, 0));
+    }
+    QCOMPARE(xDeltaSpy.size(), shouldDrag ? 1 : 0);
+    QCOMPARE(centroidChangedSpy.size(), shouldDrag ? 5 : 0);
 #if QT_CONFIG(cursor)
     QTest::mouseMove(window, p1 + QPoint(1, 0)); // TODO after fixing QTBUG-53987, don't send mouseMove
     QCOMPARE(window->cursor().shape(), Qt::ArrowCursor);
@@ -375,8 +399,9 @@ void tst_DragHandler::mouseDragThreshold()
         dragHandler->setDragThreshold(dragThreshold);
     }
 
-    QSignalSpy translationChangedSpy(dragHandler, SIGNAL(translationChanged()));
+    QSignalSpy translationChangedSpy(dragHandler, &QQuickDragHandler::translationChanged);
     QSignalSpy centroidChangedSpy(dragHandler, SIGNAL(centroidChanged()));
+    QSignalSpy xDeltaSpy(dragHandler->xAxis(), &QQuickDragAxis::activeValueChanged);
 
     QPointF ballCenter = ball->clipRect().center();
     QPointF scenePressPos = ball->mapToScene(ballCenter);
@@ -388,20 +413,28 @@ void tst_DragHandler::mouseDragThreshold()
     QCOMPARE(dragHandler->centroid().scenePosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().scenePressPosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().velocity(), QVector2D());
-    QCOMPARE(centroidChangedSpy.count(), 1);
+    QCOMPARE(centroidChangedSpy.size(), 1);
     p1 += QPoint(qMax(1, dragThreshold), 0); // QTBUG-85431: zero-distance mouse moves are not delivered
     QTest::mouseMove(window, p1);
     if (dragThreshold > 0)
         QTRY_VERIFY(dragHandler->centroid().velocity().x() > 0);
-    QCOMPARE(centroidChangedSpy.count(), 2);
+    QCOMPARE(centroidChangedSpy.size(), 2);
     // the handler is not yet active, unless the drag threshold was already exceeded
     QCOMPARE(dragHandler->active(), dragThreshold == 0);
     p1 += QPoint(1, 0);
     QTest::mouseMove(window, p1);
     QTRY_VERIFY(dragHandler->active());
-    QCOMPARE(translationChangedSpy.count(), dragThreshold ? 0 : 1);
-    QCOMPARE(centroidChangedSpy.count(), 3);
+    QCOMPARE(translationChangedSpy.size(), dragThreshold ? 0 : 1);
+    if (!dragThreshold)
+        QCOMPARE(translationChangedSpy.first().first().value<QVector2D>(), QVector2D(2, 0));
+    QCOMPARE(xDeltaSpy.size(), dragThreshold ? 0 : 1);
+    QCOMPARE(centroidChangedSpy.size(), 3);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler->translation().x(), dragThreshold ? 0 : 2);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold ? 0 : 2);
     QPointF sceneGrabPos = dragThreshold ? p1 : p1 - QPoint(1, 0);
     QCOMPARE(dragHandler->centroid().sceneGrabPosition(), sceneGrabPos);
     p1 += QPoint(19, 0);
@@ -412,16 +445,26 @@ void tst_DragHandler::mouseDragThreshold()
     QCOMPARE(dragHandler->centroid().scenePosition(), ball->mapToScene(ballCenter));
     QCOMPARE(dragHandler->centroid().scenePressPosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().sceneGrabPosition(), sceneGrabPos);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler->translation().x(), dragThreshold + (dragThreshold ? 20 : 21));
     QCOMPARE(dragHandler->translation().y(), 0.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler->activeTranslation().x(), dragThreshold + (dragThreshold ? 20 : 21));
+    QCOMPARE(dragHandler->activeTranslation().y(), 0.0);
+    QCOMPARE(translationChangedSpy.size(), dragThreshold ? 1 : 2);
+    QCOMPARE(translationChangedSpy.first().first().value<QVector2D>(),
+             QVector2D(dragThreshold ? dragThreshold + 20 : 2, 0));
     QVERIFY(dragHandler->centroid().velocity().x() > 0);
-    QCOMPARE(centroidChangedSpy.count(), 4);
+    QCOMPARE(centroidChangedSpy.size(), 4);
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, p1);
     QTRY_VERIFY(!dragHandler->active());
     QCOMPARE(dragHandler->centroid().pressedButtons(), Qt::NoButton);
     QCOMPARE(ball->mapToScene(ballCenter).toPoint(), p1);
-    QCOMPARE(translationChangedSpy.count(), dragThreshold ? 1 : 2);
-    QCOMPARE(centroidChangedSpy.count(), 5);
+    QCOMPARE(translationChangedSpy.size(), dragThreshold ? 1 : 2);
+    QCOMPARE(xDeltaSpy.size(), dragThreshold ? 1 : 2);
+    QCOMPARE(centroidChangedSpy.size(), 5);
 }
 
 void tst_DragHandler::dragFromMargin() // QTBUG-74966
@@ -451,8 +494,14 @@ void tst_DragHandler::dragFromMargin() // QTBUG-74966
     QTRY_VERIFY(dragHandler->active());
     QCOMPARE(dragHandler->centroid().scenePressPosition(), scenePressPos);
     QCOMPARE(dragHandler->centroid().sceneGrabPosition(), p1);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler->translation().x(), 0.0); // hmm that's odd
     QCOMPARE(dragHandler->translation().y(), 0.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler->activeTranslation().x(), 0.0); // hmm that's odd
+    QCOMPARE(dragHandler->activeTranslation().y(), 0.0);
     QCOMPARE(draggableItem->position(), originalPos + QPointF(dragThreshold * 2, 0));
 #if QT_CONFIG(cursor)
     // The cursor doesn't change until the next event after the handler becomes active.
@@ -551,15 +600,17 @@ void tst_DragHandler::touchDragMulti()
     QVERIFY(ball1);
     QQuickDragHandler *dragHandler1 = ball1->findChild<QQuickDragHandler*>();
     QVERIFY(dragHandler1);
-    QSignalSpy translationChangedSpy1(dragHandler1, SIGNAL(translationChanged()));
+    QSignalSpy translationChangedSpy1(dragHandler1, &QQuickDragHandler::translationChanged);
     QSignalSpy centroidChangedSpy1(dragHandler1, SIGNAL(centroidChanged()));
+    QSignalSpy xDeltaSpy1(dragHandler1->xAxis(), &QQuickDragAxis::activeValueChanged);
 
     QQuickItem *ball2 = window->rootObject()->childItems().at(1);
     QVERIFY(ball2);
     QQuickDragHandler *dragHandler2 = ball2->findChild<QQuickDragHandler*>();
     QVERIFY(dragHandler2);
-    QSignalSpy translationChangedSpy2(dragHandler2, SIGNAL(translationChanged()));
+    QSignalSpy translationChangedSpy2(dragHandler2, &QQuickDragHandler::translationChanged);
     QSignalSpy centroidChangedSpy2(dragHandler1, SIGNAL(centroidChanged()));
+    QSignalSpy yDeltaSpy2(dragHandler2->yAxis(), &QQuickDragAxis::activeValueChanged);
 
     QPointF ball1Center = ball1->clipRect().center();
     QPointF scenePressPos1 = ball1->mapToScene(ball1Center);
@@ -574,13 +625,13 @@ void tst_DragHandler::touchDragMulti()
     touchSeq.stationary(1).press(2, p2, window).commit();
     QQuickTouchUtils::flush(window);
     QVERIFY(!dragHandler1->active());
-    QCOMPARE(centroidChangedSpy1.count(), 2);
+    QCOMPARE(centroidChangedSpy1.size(), 2);
     QCOMPARE(dragHandler1->centroid().position(), ball1Center);
     QCOMPARE(dragHandler1->centroid().pressPosition(), ball1Center);
     QCOMPARE(dragHandler1->centroid().scenePosition(), scenePressPos1);
     QCOMPARE(dragHandler1->centroid().scenePressPosition(), scenePressPos1);
     QVERIFY(!dragHandler2->active());
-    QCOMPARE(centroidChangedSpy2.count(), 2);
+    QCOMPARE(centroidChangedSpy2.size(), 2);
     QCOMPARE(dragHandler2->centroid().position(), ball2Center);
     QCOMPARE(dragHandler2->centroid().pressPosition(), ball2Center);
     QCOMPARE(dragHandler2->centroid().scenePosition(), scenePressPos2);
@@ -590,13 +641,13 @@ void tst_DragHandler::touchDragMulti()
     touchSeq.move(1, p1, window).move(2, p2, window).commit();
     QQuickTouchUtils::flush(window);
     QVERIFY(!dragHandler1->active());
-    QCOMPARE(centroidChangedSpy1.count(), 3);
+    QCOMPARE(centroidChangedSpy1.size(), 3);
     QCOMPARE(dragHandler1->centroid().position(), ball1Center + QPointF(dragThreshold, 0));
     QCOMPARE(dragHandler1->centroid().pressPosition(), ball1Center);
     QCOMPARE(dragHandler1->centroid().scenePosition().toPoint(), p1);
     QCOMPARE(dragHandler1->centroid().scenePressPosition(), scenePressPos1);
     QVERIFY(!dragHandler2->active());
-    QCOMPARE(centroidChangedSpy2.count(), 3);
+    QCOMPARE(centroidChangedSpy2.size(), 3);
     QCOMPARE(dragHandler2->centroid().position(), ball2Center + QPointF(0, dragThreshold));
     QCOMPARE(dragHandler2->centroid().pressPosition(), ball2Center);
     QCOMPARE(dragHandler2->centroid().scenePosition().toPoint(), p2);
@@ -607,8 +658,14 @@ void tst_DragHandler::touchDragMulti()
     QQuickTouchUtils::flush(window);
     QTRY_VERIFY(dragHandler1->active());
     QVERIFY(dragHandler2->active());
-    QCOMPARE(translationChangedSpy1.count(), 0);
+    QCOMPARE(translationChangedSpy1.size(), 0);
+    QCOMPARE(xDeltaSpy1.size(), 0);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler1->translation().x(), 0.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler1->activeTranslation().x(), 0.0);
     QPointF sceneGrabPos1 = p1;
     QPointF sceneGrabPos2 = p2;
     QCOMPARE(dragHandler1->centroid().sceneGrabPosition(), sceneGrabPos1);
@@ -616,8 +673,14 @@ void tst_DragHandler::touchDragMulti()
     p1 += QPoint(19, 0);
     p2 += QPoint(0, 19);
     QVERIFY(dragHandler2->active());
-    QCOMPARE(translationChangedSpy2.count(), 0);
+    QCOMPARE(translationChangedSpy2.size(), 0);
+    QCOMPARE(yDeltaSpy2.size(), 0);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler2->translation().x(), 0.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler2->activeTranslation().x(), 0.0);
     QCOMPARE(dragHandler2->centroid().sceneGrabPosition(), sceneGrabPos2);
     touchSeq.move(1, p1, window).move(2, p2, window).commit();
     QQuickTouchUtils::flush(window);
@@ -628,27 +691,47 @@ void tst_DragHandler::touchDragMulti()
     QCOMPARE(dragHandler1->centroid().scenePosition(), ball1->mapToScene(ball1Center));
     QCOMPARE(dragHandler1->centroid().scenePressPosition(), scenePressPos1);
     QCOMPARE(dragHandler1->centroid().sceneGrabPosition(), sceneGrabPos1);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler1->translation().x(), dragThreshold + 20.0);
     QCOMPARE(dragHandler1->translation().y(), 0.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler1->activeTranslation().x(), dragThreshold + 20.0);
+    QCOMPARE(dragHandler1->activeTranslation().y(), 0.0);
     QCOMPARE(dragHandler2->centroid().position(), ball2Center);
     QCOMPARE(dragHandler2->centroid().pressPosition(), ball2Center);
     QCOMPARE(dragHandler2->centroid().scenePosition(), ball2->mapToScene(ball2Center));
     QCOMPARE(dragHandler2->centroid().scenePressPosition(), scenePressPos2);
     QCOMPARE(dragHandler2->centroid().sceneGrabPosition(), sceneGrabPos2);
+#if QT_DEPRECATED_SINCE(6, 2)
+QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
     QCOMPARE(dragHandler2->translation().x(), 0.0);
     QCOMPARE(dragHandler2->translation().y(), dragThreshold + 20.0);
+QT_WARNING_POP
+#endif
+    QCOMPARE(dragHandler2->activeTranslation().x(), 0.0);
+    QCOMPARE(dragHandler2->activeTranslation().y(), dragThreshold + 20.0);
+    QCOMPARE(xDeltaSpy1.size(), 1);
+    QCOMPARE(xDeltaSpy1.first().first().toReal(), dragThreshold + 20);
+    QCOMPARE(yDeltaSpy2.size(), 1);
+    QCOMPARE(yDeltaSpy2.first().first().toReal(), dragThreshold + 20);
+    QCOMPARE(translationChangedSpy1.size(), 1);
+    QCOMPARE(translationChangedSpy1.first().first().value<QVector2D>(), QVector2D(dragThreshold + 20, 0));
+    QCOMPARE(translationChangedSpy2.size(), 1);
+    QCOMPARE(translationChangedSpy2.first().first().value<QVector2D>(), QVector2D(0, dragThreshold + 20));
     touchSeq.release(1, p1, window).stationary(2).commit();
     QQuickTouchUtils::flush(window);
     QTRY_VERIFY(!dragHandler1->active());
     QVERIFY(dragHandler2->active());
     QCOMPARE(dragHandler1->centroid().pressedButtons(), Qt::NoButton);
     QCOMPARE(ball1->mapToScene(ball1Center).toPoint(), p1);
-    QCOMPARE(translationChangedSpy1.count(), 1);
+    QCOMPARE(translationChangedSpy1.size(), 1);
     touchSeq.release(2, p2, window).commit();
     QQuickTouchUtils::flush(window);
     QTRY_VERIFY(!dragHandler2->active());
     QCOMPARE(ball2->mapToScene(ball2Center).toPoint(), p2);
-    QCOMPARE(translationChangedSpy2.count(), 1);
+    QCOMPARE(translationChangedSpy2.size(), 1);
 }
 
 void tst_DragHandler::touchDragMultiSliders_data()
@@ -737,7 +820,7 @@ void tst_DragHandler::touchDragMultiSliders()
             expectedEndPosition = startPoints[sli].toPoint();
         qCDebug(lcPointerTests) << "slider " << knobs[sli]->objectName() << "started @" << startPoints[sli]
             << "tried to move by" << movements[sli] << "ended up @" << endPosition << "expected" << expectedEndPosition;
-        QTRY_COMPARE(endPosition, expectedEndPosition);
+        QCOMPARE(endPosition, expectedEndPosition);
     }
 
     // Release
@@ -854,7 +937,7 @@ void tst_DragHandler::touchPinchAndMouseMove()
     for (int i = 0; i < 10; ++i) {
         p1 += delta;
         QTest::mouseMove(window, p1);
-        QCOMPARE(rectMovedSpy.count(), 0);
+        QCOMPARE(rectMovedSpy.size(), 0);
     }
 }
 
@@ -982,6 +1065,131 @@ void tst_DragHandler::underModalLayer() // QTBUG-78258
     QTest::mouseRelease(window, Qt::LeftButton);
 }
 
+void tst_DragHandler::interruptedByIrrelevantButton() // QTBUG-102201
+{
+    const int dragThreshold = QGuiApplication::styleHints()->startDragDistance();
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("draggables.qml")));
+
+    QQuickItem *ball = window.rootObject()->childItems().first();
+    QVERIFY(ball);
+    QQuickDragHandler *dragHandler = ball->findChild<QQuickDragHandler*>();
+    QVERIFY(dragHandler);
+
+    QCOMPARE(dragHandler->acceptedButtons(), Qt::LeftButton); // the default
+
+    QSignalSpy translationChangedSpy(dragHandler, &QQuickDragHandler::translationChanged);
+    QSignalSpy cancelSpy(dragHandler, &QQuickDragHandler::canceled);
+    QSignalSpy activeSpy(dragHandler, &QQuickDragHandler::activeChanged);
+    QSignalSpy grabSpy(dragHandler, &QQuickDragHandler::grabChanged);
+
+    QPoint p1 = ball->mapToScene(ball->clipRect().center()).toPoint();
+    QTest::mousePress(&window, Qt::LeftButton, Qt::NoModifier, p1);
+    QCOMPARE(grabSpy.size(), 1); // passive grab
+    p1 += QPoint(dragThreshold + 1, 0);
+    QTest::mouseMove(&window, p1);
+    QVERIFY(dragHandler->active());
+    QCOMPARE(activeSpy.size(), 1);
+    QCOMPARE(grabSpy.size(), 2); // exclusive grab
+    QCOMPARE(translationChangedSpy.size(), 0);
+    p1 += QPoint(1, 0);
+    QTest::mouseMove(&window, p1);
+    QCOMPARE(translationChangedSpy.size(), 1);
+
+    // Left button is already held, now press right button too (chording)
+    QTest::mousePress(&window, Qt::RightButton, Qt::NoModifier, p1);
+    // DragHandler will ungrab and deactivate, but not cancel
+    QCOMPARE(dragHandler->active(), false);
+    QCOMPARE(translationChangedSpy.size(), 1);
+    QCOMPARE(cancelSpy.size(), 0);
+    QCOMPARE(activeSpy.size(), 2);
+    QCOMPARE_GT(grabSpy.size(), 2); // lost grabs
+
+    // Release right button: no change in state
+    QTest::mouseRelease(&window, Qt::RightButton, Qt::NoModifier, p1);
+    QCOMPARE(dragHandler->active(), false);
+    QCOMPARE(translationChangedSpy.size(), 1);
+    QCOMPARE(cancelSpy.size(), 0);
+    QCOMPARE(activeSpy.size(), 2);
+    auto grabChangedCount = grabSpy.size();
+
+    // But the left button is still held, and it's possible to resume dragging
+    p1 += QPoint(dragThreshold + 1, 0);
+    QTest::mouseMove(&window, p1);
+    QCOMPARE_GT(grabSpy.size(), grabChangedCount); // re-grabbed
+    p1 += QPoint(1, 0);
+    QTest::mouseMove(&window, p1);
+    QVERIFY(dragHandler->active());
+    QCOMPARE(activeSpy.size(), 3);
+    QCOMPARE(translationChangedSpy.size(), 2);
+
+    QTest::mouseRelease(&window, Qt::LeftButton, Qt::NoModifier, p1);
+    QVERIFY(!dragHandler->active());
+    QCOMPARE(activeSpy.size(), 4);
+    QCOMPARE(cancelSpy.size(), 0); // none of this caused a canceled() signal
+}
+
+class TouchItem : public QQuickItem {
+public:
+    TouchItem(QQuickItem *parent = nullptr) : QQuickItem(parent)
+    {
+        setAcceptTouchEvents(true);
+    }
+
+protected:
+    void touchEvent(QTouchEvent *ev) override
+    {
+        switch (ev->type()) {
+        case QEvent::TouchBegin:
+            ev->accept();
+            break;
+        default:
+            break;
+        }
+    }
+};
+void tst_DragHandler::touchDragExclusiveGrabber()
+{
+    qmlRegisterType<TouchItem>("Test", 1, 0, "TouchItem");
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("dragHandlerTakeOverForbidden.qml")));
+
+    QQuickItem *rect = window.rootObject()->findChild<QQuickItem *>();
+    QVERIFY(rect);
+    QQuickDragHandler *dragHandler = rect->findChild<QQuickDragHandler*>();
+    QVERIFY(dragHandler);
+
+    QSignalSpy centroidChangedSpy(dragHandler, &QQuickDragHandler::centroidChanged);
+    QSignalSpy dragHandlerActiveChangedSpy(dragHandler, &QQuickDragHandler::activeChanged);
+    QSignalSpy grabChangedSpy(dragHandler, &QQuickDragHandler::grabChanged);
+
+    QPointF rectCenter = rect->clipRect().center();
+    QPointF scenePressPos = rect->mapToScene(rectCenter);
+    QPoint p1 = scenePressPos.toPoint();
+    auto dragThreshold = dragHandler->dragThreshold();
+
+    QTest::QTouchEventSequence touchSeq = QTest::touchEvent(&window, touchDevice, false);
+    touchSeq.press(1, p1, &window).commit();
+    QVERIFY(!dragHandler->active());
+    QCOMPARE(dragHandler->centroid().velocity(), QVector2D());
+    QCOMPARE(centroidChangedSpy.size(), 1);
+    QCOMPARE(grabChangedSpy.size(), 1);
+
+    p1 += QPoint(dragThreshold + 1, dragThreshold + 1);
+    touchSeq.move(1, p1, &window).commit();
+    QQuickTouchUtils::flush(&window);
+    QCOMPARE(dragHandler->active(), false); // TakeOverForbidden so approveGrabTransition says no
+    QCOMPARE(centroidChangedSpy.size(), 1);
+    QCOMPARE(grabChangedSpy.size(), 1);
+    QCOMPARE(grabChangedSpy.takeLast().at(0).toInt(), QPointingDevice::GrabTransition::GrabPassive);
+
+    touchSeq.release(1, p1, &window).commit();
+    QQuickTouchUtils::flush(&window);
+    QVERIFY(!dragHandler->active());
+    QCOMPARE(centroidChangedSpy.size(), 1);
+    QCOMPARE(grabChangedSpy.size(), 1);
+}
+
 QTEST_MAIN(tst_DragHandler)
 
 #include "tst_qquickdraghandler.moc"
