@@ -1,37 +1,14 @@
-/****************************************************************************
-**
-** Copyright (C) 2020 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2020 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <qtest.h>
 
+#include <QtQml/QQmlComponent>
 #include <QtQuick/qquickitem.h>
 #include <QtQuick/qquickwindow.h>
 #include <QtQuick/qquickview.h>
 #include "private/qquickfocusscope_p.h"
+#include "private/qquickrectangle_p.h"
 #include "private/qquickitem_p.h"
 #include <QtGui/private/qevent_p.h>
 #include <qpa/qwindowsysteminterface.h>
@@ -47,7 +24,7 @@
 #include <QTranslator>
 #include <QtCore/qregularexpression.h>
 
-#ifdef TEST_QTBUG_60123
+#ifdef QT_WIDGETS_LIB
 #include <QWidget>
 #include <QMainWindow>
 #endif
@@ -240,7 +217,7 @@ private slots:
 
     void shortcutOverride();
 
-#ifdef TEST_QTBUG_60123
+#ifdef QT_WIDGETS_LIB
     void qtBug60123();
 #endif
 
@@ -250,6 +227,11 @@ private slots:
     void polishLoopDetection_data();
     void polishLoopDetection();
 
+    void objectCastInDestructor();
+    void listsAreNotLists();
+
+    void transformChanged();
+
 private:
 
     enum PaintOrderOp {
@@ -1059,13 +1041,13 @@ void tst_qquickitem::constructor()
     QQuickItem *child1 = new QQuickItem(root.data());
     QCOMPARE(child1->parent(), root.data());
     QCOMPARE(child1->parentItem(), root.data());
-    QCOMPARE(root->childItems().count(), 1);
+    QCOMPARE(root->childItems().size(), 1);
     QCOMPARE(root->childItems().at(0), child1);
 
     QQuickItem *child2 = new QQuickItem(root.data());
     QCOMPARE(child2->parent(), root.data());
     QCOMPARE(child2->parentItem(), root.data());
-    QCOMPARE(root->childItems().count(), 2);
+    QCOMPARE(root->childItems().size(), 2);
     QCOMPARE(root->childItems().at(0), child1);
     QCOMPARE(root->childItems().at(1), child2);
 }
@@ -1083,7 +1065,7 @@ void tst_qquickitem::setParentItem()
     child1->setParentItem(root);
     QVERIFY(!child1->parent());
     QCOMPARE(child1->parentItem(), root);
-    QCOMPARE(root->childItems().count(), 1);
+    QCOMPARE(root->childItems().size(), 1);
     QCOMPARE(root->childItems().at(0), child1);
 
     QQuickItem *child2 = new QQuickItem;
@@ -1092,14 +1074,14 @@ void tst_qquickitem::setParentItem()
     child2->setParentItem(root);
     QVERIFY(!child2->parent());
     QCOMPARE(child2->parentItem(), root);
-    QCOMPARE(root->childItems().count(), 2);
+    QCOMPARE(root->childItems().size(), 2);
     QCOMPARE(root->childItems().at(0), child1);
     QCOMPARE(root->childItems().at(1), child2);
 
     child1->setParentItem(nullptr);
     QVERIFY(!child1->parent());
     QVERIFY(!child1->parentItem());
-    QCOMPARE(root->childItems().count(), 1);
+    QCOMPARE(root->childItems().size(), 1);
     QCOMPARE(root->childItems().at(0), child2);
 
     delete root;
@@ -1115,7 +1097,9 @@ void tst_qquickitem::setParentItem()
 
 void tst_qquickitem::visible()
 {
+    QQuickWindow window;
     QQuickItem *root = new QQuickItem;
+    root->setParentItem(window.contentItem());
 
     QQuickItem *child1 = new QQuickItem;
     child1->setParentItem(root);
@@ -1545,7 +1529,7 @@ void tst_qquickitem::polishLoopDetection_data()
 
     QTest::newRow("test1.100") <<   PolishItemSpans({ {1, 100} }) << 0;
     QTest::newRow("test1.1002") <<  PolishItemSpans({ {1, 1002} }) << 3;
-    QTest::newRow("test1.2020") <<  PolishItemSpans({ {1, 2020} }) << 10;
+    QTest::newRow("test1.2020") <<  PolishItemSpans({ {1, 2020} }) << 5;
 
     QTest::newRow("test5.1") <<    PolishItemSpans({ {5, 1} }) << 0;
     QTest::newRow("test5.10") <<   PolishItemSpans({ {5, 10} }) << 0;
@@ -1589,7 +1573,7 @@ void tst_qquickitem::polishLoopDetection()
     }
 
     QList<QQuickItem*> items = window.contentItem()->childItems();
-    for (int i = 0; i < items.count(); ++i) {
+    for (int i = 0; i < items.size(); ++i) {
         static_cast<TestPolishItem*>(items.at(i))->doPolish();
     }
     item = static_cast<TestPolishItem*>(items.first());
@@ -1945,7 +1929,7 @@ void tst_qquickitem::paintOrder()
     QList<QQuickItem*> list = QQuickItemPrivate::get(root)->paintOrderChildItems();
 
     QStringList items;
-    for (int i = 0; i < list.count(); ++i)
+    for (int i = 0; i < list.size(); ++i)
         items << list.at(i)->objectName();
 
     QCOMPARE(items, expected);
@@ -2012,13 +1996,6 @@ void tst_qquickitem::acceptedMouseButtons()
     QCOMPARE(item.releaseCount, 3);
 }
 
-static void gc(QQmlEngine &engine)
-{
-    engine.collectGarbage();
-    QCoreApplication::sendPostedEvents(nullptr, QEvent::DeferredDelete);
-    QCoreApplication::processEvents();
-}
-
 void tst_qquickitem::visualParentOwnership()
 {
     QQmlEngine engine;
@@ -2357,7 +2334,7 @@ void tst_qquickitem::shortcutOverride()
     QCOMPARE(view.rootObject()->property("shortcutActivationCount").toInt(), 1);
 }
 
-#ifdef TEST_QTBUG_60123
+#ifdef QT_WIDGETS_LIB
 void tst_qquickitem::qtBug60123()
 {
     QMainWindow main;
@@ -2459,6 +2436,247 @@ void tst_qquickitem::receivesLocaleChangeEvent()
     QCOMPARE(child2->localeChangeEventCount, 1);
 }
 
+void tst_qquickitem::objectCastInDestructor()
+{
+    QQuickView view;
+    view.setSource(testFileUrl("objectCastInDestructor.qml"));
+    view.show();
+
+    QQuickItem *item = view.findChild<QQuickItem *>("testRectangle");
+    QVERIFY(item);
+    bool destroyed = false;
+    connect(item, &QObject::destroyed, [&]{
+        destroyed = true;
+        QCOMPARE(qobject_cast<QQuickItem *>(item), nullptr);
+        QCOMPARE(qobject_cast<QQuickRectangle *>(item), nullptr);
+    });
+
+    QQuickItem *loader = view.findChild<QQuickItem *>("loader");
+    QVERIFY(loader);
+    loader->setProperty("active", false);
+
+    QVERIFY(QTest::qWaitFor([&destroyed]{ return destroyed; }));
+}
+
+template<typename T>
+void verifyListProperty(const T &data)
+{
+    QVERIFY(data.object);
+    QVERIFY(data.append);
+    QVERIFY(data.count);
+    QVERIFY(data.at);
+    QVERIFY(data.clear);
+    QVERIFY(data.removeLast);
+
+    // We must not synthesize the replace and removeLast methods on those properties.
+    // They would be even more broken than the explicitly defined methods.
+    QVERIFY(!data.replace);
+}
+
+void tst_qquickitem::listsAreNotLists()
+{
+    QQuickItem item;
+    QQuickItem child;
+    QObject resource;
+
+    QQmlListProperty<QObject> data
+        = item.property("data").value<QQmlListProperty<QObject>>();
+    QQmlListProperty<QObject> resources
+        = item.property("resources").value<QQmlListProperty<QObject>>();
+    QQmlListProperty<QQuickItem> children
+        = item.property("children").value<QQmlListProperty<QQuickItem>>();
+
+    verifyListProperty(data);
+    verifyListProperty(resources);
+    verifyListProperty(children);
+
+
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    children.append(&children, &child);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 1);
+    children.removeLast(&children);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    data.append(&data, &child);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 1);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    children.append(&children, &child);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 1);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    data.append(&data, &child);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 1);
+    children.removeLast(&children);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+
+
+    resources.append(&resources, &resource);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    resources.removeLast(&resources);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    data.append(&data, &resource);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    resources.append(&resources, &resource);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+    data.append(&data, &resource);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    resources.removeLast(&resources);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+
+
+    children.append(&children, &child);
+    resources.append(&resources, &resource);
+    QCOMPARE(data.count(&data), 2);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 1);
+    children.removeLast(&children);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    resources.removeLast(&resources);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+
+
+    children.append(&children, &child);
+    resources.append(&resources, &resource);
+    QCOMPARE(data.count(&data), 2);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 1);
+    resources.removeLast(&resources);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 1);
+    children.removeLast(&children);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+
+
+    data.append(&data, &child);
+    data.append(&data, &resource);
+    QCOMPARE(data.count(&data), 2);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 1);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 1);
+    QCOMPARE(resources.count(&resources), 1);
+    QCOMPARE(children.count(&children), 0);
+    data.removeLast(&data);
+    QCOMPARE(data.count(&data), 0);
+    QCOMPARE(resources.count(&resources), 0);
+    QCOMPARE(children.count(&children), 0);
+}
+
+class TransformItemPrivate;
+class TransformItem : public QQuickItem {
+    Q_OBJECT
+    Q_DECLARE_PRIVATE(TransformItem)
+public:
+    TransformItem(QQuickItem *parent = nullptr);
+
+    bool transformChanged = false;
+};
+
+class TransformItemPrivate :public QQuickItemPrivate
+{
+protected:
+    Q_DECLARE_PUBLIC(TransformItem)
+
+    bool transformChanged(QQuickItem *) override
+    {
+        Q_Q(TransformItem);
+        q->transformChanged = true;
+        return true;
+    }
+};
+
+TransformItem::TransformItem(QQuickItem *parent)
+    : QQuickItem(*(new TransformItemPrivate), parent)
+{
+}
+
+void tst_qquickitem::transformChanged()
+{
+    QQuickItem rootItem;
+    QQuickItem *parents[2][3] = {};
+    for (int i = 0; i < 2; ++i) {
+        for (int j = 0; j < 3; ++j) {
+            auto *item = new QQuickItem;
+            item->setObjectName(QString("Item-%1.%2").arg(i).arg(j));
+            item->setParentItem(j > 0 ? parents[i][j - 1] : &rootItem);
+            parents[i][j] = item;
+        }
+    }
+
+    // Setting the root item's position will result in transformChanged,
+    // which will clear the subtreeTransformChangedEnabled flag that is
+    // by default set to true, since there are no children that observe
+    // the viewport.
+    parents[0][0]->setPosition(QPoint(100, 100));
+    parents[1][0]->setPosition(QPoint(200, 200));
+
+    TransformItem transformItem;
+    transformItem.setParentItem(parents[0][2]);
+    transformItem.setFlag(QQuickItem::ItemObservesViewport);
+    QCOMPARE(transformItem.mapToScene(QPoint(0, 0)), parents[0][0]->position());
+
+    parents[0][0]->setPosition(QPoint(110, 110));
+    QVERIFY2(transformItem.transformChanged,
+        "Moving an ancestor should trigger transformChanged");
+
+    transformItem.transformChanged = false;
+    transformItem.setParentItem(parents[1][2]);
+    QVERIFY2(transformItem.transformChanged,
+        "Reparenting the item should result in a transformChanged");
+    QCOMPARE(transformItem.mapToScene(QPoint(0, 0)), parents[1][0]->position());
+
+    transformItem.transformChanged = false;
+    parents[1][0]->setPosition(QPoint(220, 220));
+    QVERIFY2(transformItem.transformChanged,
+        "Changing one of the new ancestors should result in transformChanged");
+    QCOMPARE(transformItem.mapToScene(QPoint(0, 0)), parents[1][0]->position());
+}
+
 QTEST_MAIN(tst_qquickitem)
 
 #include "tst_qquickitem.moc"
