@@ -1,7 +1,10 @@
 // Copyright (C) 2021 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
-#include <QtTest/QtTest>
+#include <QtTest/QTest>
+#include <QtTest/QSignalSpy>
+#include <QtCore/QSortFilterProxyModel>
+#include <QtCore/QStringListModel>
 #include <QtQuickTest/quicktest.h>
 
 #include <QtQuick/qquickview.h>
