@@ -1,9 +1,10 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <qtest.h>
 #include <qtesttouch.h>
 #include <QtTest/QSignalSpy>
+#include <QtTest/QTestAccessibility>
 #include <QtTest/private/qtesthelpers_p.h>
 #include <QtQml/qqmlcomponent.h>
 #include <QtQml/qqmlcontext.h>
@@ -13,10 +14,15 @@
 #include <QtQuick/private/qquickmousearea_p.h>
 #include <QtQuick/private/qquicktaphandler_p.h>
 #include <QtQuickTemplates2/private/qquickbutton_p.h>
+#include <QtQuickTemplates2/private/qquickoverlay_p.h>
+#include <QtQuickTemplates2/private/qquickpopup_p.h>
 #include <QtQuickTestUtils/private/qmlutils_p.h>
+#include <QtQuickTestUtils/private/visualtestutils_p.h>
 #include <QtGui/QWindow>
 #include <QtGui/QScreen>
+#include <QtGui/QAccessible>
 #include <QtGui/QImage>
+#include <QtGui/qpa/qplatformaccessibility.h>
 #include <QtCore/QDebug>
 #include <QtQml/qqmlengine.h>
 
@@ -28,6 +34,8 @@
 
 #include <QtQuickWidgets/QQuickWidget>
 
+using namespace Qt::StringLiterals;
+
 #if QT_CONFIG(graphicsview)
 # include <QtWidgets/QGraphicsView>
 # include <QtWidgets/QGraphicsProxyWidget>
@@ -107,6 +115,9 @@ class tst_qquickwidget : public QQmlDataTest
 public:
     tst_qquickwidget();
 
+private:
+    bool initAccessibility();
+
 private slots:
     void showHide();
     void reparentAfterShow();
@@ -133,7 +144,11 @@ private slots:
     void touchTapHandler();
     void touchMultipleWidgets();
     void tabKey();
+    void focusChain_data();
+    void focusChain();
     void resizeOverlay();
+    void overlayGeometry_data();
+    void overlayGeometry();
     void controls();
     void focusOnClick();
 #if QT_CONFIG(graphicsview)
@@ -141,7 +156,13 @@ private slots:
 #endif
     void focusPreserved();
     void accessibilityHandlesViewChange();
+    void accessibleParentOfQuickItems();
     void cleanupRhi();
+    void dontRecreateRootElementOnWindowChange();
+    void setInitialProperties();
+    void fromModuleCtor();
+    void loadFromModule_data();
+    void loadFromModule();
 
 private:
     QPointingDevice *device = QTest::createTouchDevice();
@@ -153,6 +174,17 @@ tst_qquickwidget::tst_qquickwidget()
 {
 }
 
+bool tst_qquickwidget::initAccessibility()
+{
+    // Copied from tst_QQuickAccessible::initTestCase()
+    QTestAccessibility::initialize();
+    QPlatformIntegration *pfIntegration = QGuiApplicationPrivate::platformIntegration();
+    if (!pfIntegration->accessibility())
+        return false;
+    pfIntegration->accessibility()->setActive(true);
+    return true;
+}
+
 void tst_qquickwidget::showHide()
 {
     QWidget window;
@@ -220,6 +252,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(view->size(), QSize(200, 200));
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->size(), view->initialSize());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from view
     view->resize(QSize(80,100));
@@ -228,6 +261,8 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 100.0);
     QCOMPARE(view->size(), QSize(80, 100));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 
     view->setResizeMode(QQuickWidget::SizeViewToRootObject);
 
@@ -236,6 +271,8 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->width(), 80.0);
     QCOMPARE(item->height(), 100.0);
     QTRY_COMPARE(view->size(), QSize(60, 80));
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 
     // size update from root object
     item->setWidth(250);
@@ -245,6 +282,7 @@ void tst_qquickwidget::resizemodeitem()
     QTRY_COMPARE(view->size(), QSize(250, 350));
     QCOMPARE(view->size(), QSize(250, 350));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // reset window
     window.hide();
@@ -262,6 +300,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 200.0);
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->size(), view->initialSize());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from root object
     item->setWidth(80);
@@ -270,6 +309,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 100.0);
     QTRY_COMPARE(view->size(), QSize(80, 100));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from root object disabled
     view->setResizeMode(QQuickWidget::SizeRootObjectToView);
@@ -278,6 +318,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(view->width(), 80);
     QCOMPARE(view->height(), 100);
     QCOMPARE(QSize(item->width(), item->height()), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     // size update from view
     view->resize(QSize(200,300));
@@ -285,6 +326,7 @@ void tst_qquickwidget::resizemodeitem()
     QCOMPARE(item->height(), 300.0);
     QCOMPARE(view->size(), QSize(200, 300));
     QCOMPARE(view->size(), view->sizeHint());
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
 
     window.hide();
 
@@ -308,6 +350,8 @@ void tst_qquickwidget::resizemodeitem()
     QTRY_COMPARE(view->size(), QSize(300, 300));
     QCOMPARE(view->size(), view->sizeHint());
     QCOMPARE(view->initialSize(), QSize(200, 200)); // initial object size
+    QCOMPARE(view->quickWindow()->contentItem()->size(), view->size());
+
 }
 
 void tst_qquickwidget::layoutSizeChange()
@@ -703,7 +747,9 @@ void tst_qquickwidget::touchTapHandler()
 
     QCoreApplication::setAttribute(Qt::AA_SynthesizeMouseForUnhandledTouchEvents, guiSynthMouse);
     QQuickWidget quick;
-    QVERIFY(quick.testAttribute(Qt::WA_AcceptTouchEvents));
+    if (!quick.testAttribute(Qt::WA_AcceptTouchEvents))
+        QSKIP("irrelevant on non-touch platforms");
+
     quick.setSource(testFileUrl("tapHandler.qml"));
     quick.show();
     QVERIFY(QTest::qWaitForWindowExposed(&quick));
@@ -728,6 +774,9 @@ void tst_qquickwidget::touchMultipleWidgets()
     QWidget window;
     QQuickWidget *leftQuick = new QQuickWidget;
     leftQuick->setSource(testFileUrl("button.qml"));
+    if (!leftQuick->testAttribute(Qt::WA_AcceptTouchEvents))
+        QSKIP("irrelevant on non-touch platforms");
+
     QQuickWidget *rightQuick = new QQuickWidget;
     rightQuick->setSource(testFileUrl("button.qml"));
 
@@ -774,9 +823,9 @@ void tst_qquickwidget::tabKey()
     qqw2->setSource(testFileUrl("noActiveFocusOnTab.qml"));
     qqw2->move(100, 0);
     window1.show();
-    qqw->setFocus();
     QVERIFY(QTest::qWaitForWindowExposed(&window1, 5000));
-    QVERIFY(qqw->hasFocus());
+    qqw->setFocus();
+    QTRY_VERIFY(qqw->hasFocus());
     QQuickItem *item = qobject_cast<QQuickItem *>(qqw->rootObject());
     QQuickItem *topItem = item->findChild<QQuickItem *>("topRect");
     QQuickItem *middleItem = item->findChild<QQuickItem *>("middleRect");
@@ -797,7 +846,98 @@ void tst_qquickwidget::tabKey()
     QVERIFY(topItem2->property("activeFocus").toBool());
     QTest::keyClick(qqw2, Qt::Key_Tab);
     QTRY_VERIFY(qqw->hasFocus());
-    QVERIFY(middleItem->property("activeFocus").toBool());
+    // Focus-in by tab/backtab events makes the first/last focusable item to be focused
+    QVERIFY(topItem->property("activeFocus").toBool());
+}
+
+void tst_qquickwidget::focusChain_data()
+{
+    if (QGuiApplication::styleHints()->tabFocusBehavior() != Qt::TabFocusAllControls)
+        QSKIP("This function doesn't support NOT iterating all.");
+
+    QTest::addColumn<bool>("forward");
+
+    QTest::addRow("Forward") << true;
+    QTest::addRow("Backward") << false;
+}
+
+void tst_qquickwidget::focusChain()
+{
+    QFETCH(bool, forward);
+
+    QWidget window;
+    window.resize(300, 300);
+    QHBoxLayout layout(&window);
+    QQuickWidget qqw1;
+    qqw1.setObjectName("qqw1");
+    qqw1.setResizeMode(QQuickWidget::SizeRootObjectToView);
+    qqw1.setSource(testFileUrl("activeFocusOnTab.qml"));
+    QWidget middleWidget;
+    middleWidget.setObjectName("middleWidget");
+    middleWidget.setFocusPolicy(Qt::FocusPolicy::TabFocus);
+    QQuickWidget qqw2;
+    qqw2.setObjectName("qqw2");
+    qqw1.setResizeMode(QQuickWidget::SizeRootObjectToView);
+    qqw2.setSource(testFileUrl("activeFocusOnTab.qml"));
+    layout.addWidget(&qqw1);
+    layout.addWidget(&middleWidget);
+    layout.addWidget(&qqw2);
+    window.show();
+    window.windowHandle()->requestActivate();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickItem *root1 = qqw1.rootObject();
+    QVERIFY(root1);
+    QQuickItem *topItem1 = root1->findChild<QQuickItem *>("topRect");
+    topItem1->setObjectName("topRect1");
+    QVERIFY(topItem1);
+    QQuickItem *middleItem1 = root1->findChild<QQuickItem *>("middleRect");
+    middleItem1->setObjectName("middleRect1");
+    QVERIFY(middleItem1);
+    QQuickItem *bottomItem1 = root1->findChild<QQuickItem *>("bottomRect");
+    bottomItem1->setObjectName("bottomRect1");
+    QVERIFY(bottomItem1);
+
+    QQuickItem *root2 = qqw2.rootObject();
+    QVERIFY(root2);
+    QQuickItem *topItem2 = root2->findChild<QQuickItem *>("topRect");
+    topItem2->setObjectName("topRect2");
+    QVERIFY(topItem2);
+    QQuickItem *middleItem2 = root2->findChild<QQuickItem *>("middleRect");
+    middleItem2->setObjectName("middleRect2");
+    QVERIFY(middleItem2);
+    QQuickItem *bottomItem2 = root2->findChild<QQuickItem *>("bottomRect");
+    bottomItem2->setObjectName("bottomRect2");
+    QVERIFY(bottomItem2);
+
+    qqw1.setFocus();
+    QTRY_VERIFY(qqw1.hasFocus());
+
+    auto hasActiveFocus = [](const QObject *targetObject) -> bool
+    {
+        if (const QQuickItem *targetItem = qobject_cast<const QQuickItem *>(targetObject))
+            return targetItem->hasActiveFocus();
+        else if (const QWidget *targetWidget = qobject_cast<const QWidget *>(targetObject))
+            return targetWidget->hasFocus();
+        return false;
+    };
+
+    QList<QObject *> expectedFocusChain;
+    if (forward) {
+        expectedFocusChain << topItem1 << middleItem1 << bottomItem1 << &middleWidget << topItem2
+                           << middleItem2 << bottomItem2 << topItem1;
+    } else {
+        expectedFocusChain << topItem1 << bottomItem2 << middleItem2 << topItem2 << &middleWidget
+                           << bottomItem1 << middleItem1 << topItem1;
+    }
+
+    const Qt::Key tabKey = forward ? Qt::Key_Tab : Qt::Key_Backtab;
+
+    for (QObject *expectedFocusTarget : expectedFocusChain) {
+        const QByteArray objectName = expectedFocusTarget->objectName().toLatin1();
+        QTRY_VERIFY2(hasActiveFocus(expectedFocusTarget), "expectedFocusTarget: " + objectName);
+        QTest::keyClick(QGuiApplication::focusWindow(), tabKey);
+    }
 }
 
 class Overlay : public QQuickItem, public QQuickItemChangeListener
@@ -839,7 +979,7 @@ void tst_qquickwidget::resizeOverlay()
     auto contentVerticalLayout = new QVBoxLayout(&widget);
     contentVerticalLayout->setContentsMargins(0, 0, 0, 0);
 
-    qmlRegisterType<Overlay>("Test", 1, 0, "Overlay");
+    qmlRegisterType<Overlay>("TestModule", 1, 0, "Overlay");
 
     auto quickWidget = new QQuickWidget(testFileUrl("resizeOverlay.qml"), &widget);
     QCOMPARE(quickWidget->status(), QQuickWidget::Ready);
@@ -868,6 +1008,55 @@ void tst_qquickwidget::resizeOverlay()
     QCOMPARE(overlay->height(), rootItem->height());
 }
 
+/*
+    Overlaps with resizeOverlay, but uses a proper
+    Qt Quick Controls Overlay.
+*/
+void tst_qquickwidget::overlayGeometry_data()
+{
+    QTest::addColumn<QQuickWidget::ResizeMode>("resizeMode");
+
+    QTest::addRow("SizeRootObjectToView") << QQuickWidget::SizeRootObjectToView;
+    QTest::addRow("SizeViewToRootObject") << QQuickWidget::SizeViewToRootObject;
+}
+
+void tst_qquickwidget::overlayGeometry()
+{
+    QFETCH(const QQuickWidget::ResizeMode, resizeMode);
+
+    QWidget window;
+
+    QQuickWidget widget(testFileUrl("overlayGeometry.qml"), &window);
+    widget.setResizeMode(resizeMode);
+
+    QCOMPARE(widget.status(), QQuickWidget::Ready);
+    const auto *rootItem = qobject_cast<QQuickItem*>(widget.rootObject());
+    QVERIFY(rootItem);
+    const QSizeF preferredSize = rootItem->size();
+
+    auto *popup = rootItem->findChild<QQuickPopup *>("popup");
+    QVERIFY(popup);
+    popup->open();
+    QTRY_VERIFY(popup->isOpened());
+
+    const auto *overlay = QQuickOverlay::overlay(widget.quickWindow());
+    QVERIFY(overlay);
+
+    QTestPrivate::androidCompatibleShow(&window);
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QCOMPARE(widget.size(), preferredSize.toSize());
+    QCOMPARE(overlay->position(), QPointF(0, 0));
+    QCOMPARE(overlay->size(), widget.size());
+
+    widget.resize((preferredSize * 2).toSize());
+    QCOMPARE(overlay->position(), QPointF(0, 0));
+    QCOMPARE(overlay->size(), widget.size());
+
+    // click outside the popup, make sure it closes
+    QTest::mouseClick(&widget, Qt::LeftButton, {}, QPoint(10, 10));
+    QTRY_VERIFY(!popup->isOpened());
+}
+
 void tst_qquickwidget::controls()
 {
     // Smoke test for having some basic Quick Controls in a scene in a QQuickWidget.
@@ -986,8 +1175,7 @@ void tst_qquickwidget::focusOnClickInProxyWidget()
 
 void tst_qquickwidget::focusPreserved()
 {
-    if (!QGuiApplicationPrivate::platformIntegration()->hasCapability(QPlatformIntegration::WindowActivation))
-        QSKIP("Window Activation is not supported.");
+    SKIP_IF_NO_WINDOW_ACTIVATION;
     if (QGuiApplication::platformName() == "android")
         QSKIP("Test doesn't exit cleanly on Android and generates many warnings - QTBUG-112696");
 
@@ -1002,6 +1190,16 @@ void tst_qquickwidget::focusPreserved()
     root->setFlag(QQuickItem::ItemHasContents);
     content->setParentItem(root);
 
+    // added content2 and content3 to test more complicated case
+    QScopedPointer<QQuickItem> content2(new QQuickItem());
+    content2->setActiveFocusOnTab(true);
+    content2->setFocus(true);
+    content2->setParentItem(root);
+    QScopedPointer<QQuickItem> content3(new QQuickItem());
+    content3->setActiveFocusOnTab(true);
+    content3->setFocus(true);
+    content3->setParentItem(root);
+
     quick->setGeometry(0, 0, 200, 200);
     quick->show();
     quick->setFocus();
@@ -1011,6 +1209,10 @@ void tst_qquickwidget::focusPreserved()
     QTRY_VERIFY(content->hasFocus());
     QTRY_VERIFY(content->hasActiveFocus());
 
+    content2->forceActiveFocus();
+    QVERIFY(content2->hasFocus());
+    QVERIFY(content2->hasActiveFocus());
+
     widget->show();
     widget->setFocus();
     widget->activateWindow();
@@ -1023,8 +1225,8 @@ void tst_qquickwidget::focusPreserved()
     quick->setFocus();
     quick->activateWindow();
     QTRY_VERIFY(quick->hasFocus());
-    QTRY_VERIFY(content->hasFocus());
-    QTRY_VERIFY(content->hasActiveFocus());
+    QTRY_VERIFY(content2->hasFocus());
+    QTRY_VERIFY(content2->hasActiveFocus());
 }
 
 /*
@@ -1069,6 +1271,47 @@ void tst_qquickwidget::accessibilityHandlesViewChange()
     (void)iface->child(0);
 }
 
+/*
+QTBUG-130116
+Since the quick widget reports the quick items as its children, then
+the quick items should report the quick widget as their parent.
+*/
+void tst_qquickwidget::accessibleParentOfQuickItems()
+{
+    if (!initAccessibility())
+        QSKIP("This platform does not support accessibility");
+
+    // These lines are copied from accessibilityHandlesViewChange()
+    if (QGuiApplication::platformName() == "offscreen")
+        QSKIP("Doesn't test anything on offscreen platform.");
+
+    QWidget window;
+    window.resize(300, 300);
+
+    QQuickWidget *quickWidget = new QQuickWidget(&window);
+    quickWidget->setSource(testFileUrl("button.qml"));
+
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    auto *rootItem = quickWidget->rootObject();
+    QVERIFY(rootItem);
+
+    QQuickButton *button = rootItem->findChild<QQuickButton *>("button");
+    QVERIFY(button);
+
+    QAccessibleInterface *iface_quickWidget = QAccessible::queryAccessibleInterface(quickWidget);
+    QVERIFY(iface_quickWidget);
+
+    QAccessibleInterface *iface_button = QAccessible::queryAccessibleInterface(button);
+    QVERIFY(iface_button);
+
+    QAccessibleInterface *iface_popup = iface_button->parent();
+    QVERIFY(iface_popup);
+    QVERIFY(iface_popup->parent());
+    QCOMPARE(iface_popup->parent(), iface_quickWidget);
+}
+
 class CreateDestroyWidget : public QWidget
 {
 public:
@@ -1078,9 +1321,6 @@ public:
 
 void tst_qquickwidget::cleanupRhi()
 {
-#ifdef Q_OS_ANDROID
-    QSKIP("This test crashes on Android (QTBUG-121133)");
-#endif
     CreateDestroyWidget topLevel;
     QQuickWidget quickWidget(&topLevel);
     quickWidget.setSource(testFileUrl("rectangle.qml"));
@@ -1091,6 +1331,77 @@ void tst_qquickwidget::cleanupRhi()
     topLevel.create();
 }
 
+void tst_qquickwidget::dontRecreateRootElementOnWindowChange()
+{
+    auto *quickWidget = new QQuickWidget();
+    quickWidget->setSource(testFileUrl("rectangle.qml"));
+    QObject *item = quickWidget->rootObject();
+
+    bool wasDestroyed = false;
+    QObject::connect(item, &QObject::destroyed, this, [&] { wasDestroyed = true; });
+
+    QEvent event(QEvent::WindowChangeInternal);
+    QCoreApplication::sendEvent(quickWidget, &event);
+
+    QVERIFY(!wasDestroyed);
+}
+
+void tst_qquickwidget::setInitialProperties()
+{
+    QQuickWidget widget;
+    widget.setInitialProperties({{"z", 4}, {"width", 100}});
+    widget.setSource(testFileUrl("resizemodeitem.qml"));
+    widget.show();
+    QObject *rootObject = widget.rootObject();
+    QVERIFY(rootObject);
+    QCOMPARE(rootObject->property("z").toInt(), 4);
+    QCOMPARE(rootObject->property("width").toInt(), 100);
+}
+
+void tst_qquickwidget::fromModuleCtor()
+{
+    QQuickWidget widget("QtQuick", "Rectangle");
+    // creation is always synchronous for C++ defined types, so we don't need _TRY
+    QObject *rootObject = widget.rootObject();
+    QVERIFY(rootObject);
+    QCOMPARE(rootObject->metaObject()->className(), "QQuickRectangle");
+}
+
+void tst_qquickwidget::loadFromModule_data()
+{
+    QTest::addColumn<QString>("module");
+    QTest::addColumn<QString>("typeName");
+    QTest::addColumn<QUrl>("url");
+    QTest::addColumn<QQuickWidget::Status>("status");
+
+    QTest::addRow("Item") << u"QtQuick"_s << u"Item"_s << QUrl() << QQuickWidget::Ready;
+    QTest::addRow("composite") << u"test"_s << u"TestQml"_s << QUrl("qrc:/qt/qml/test/data/TestQml.qml") << QQuickWidget::Ready;
+    QTest::addRow("nonexistent") << u"missing"_s << u"Type"_s << QUrl() << QQuickWidget::Error;
+}
+
+void tst_qquickwidget::loadFromModule()
+{
+    QFETCH(QString, module);
+    QFETCH(QString, typeName);
+    QFETCH(QUrl, url);
+    QFETCH(QQuickWidget::Status, status);
+
+    QQuickWidget widget;
+    widget.loadFromModule(module, typeName);
+    QTRY_COMPARE(widget.status(), status);
+    QCOMPARE(widget.source(), url);
+    if (status == QQuickWidget::Ready) {
+        QPointer<QObject> rootObject = widget.rootObject();
+        QVERIFY(rootObject);
+        // loadFromModule sets the source and deletes
+        // any object that was previously created
+        widget.loadFromModule("QtTest", "SignalSpy");
+        QVERIFY(rootObject.isNull());
+        QCOMPARE(widget.status(), QQuickWidget::Ready);
+        QVERIFY(widget.source().toString().endsWith("SignalSpy.qml"));
+    }
+}
+
 QTEST_MAIN(tst_qquickwidget)
 
 #include "tst_qquickwidget.moc"
